from copy import deepcopy

from worker.data.serialization import serialization
from worker.wellbore.model.hole_section import HoleSection


@serialization
class Riser(HoleSection):
    SERIALIZED_VARIABLES = deepcopy(HoleSection.SERIALIZED_VARIABLES)
    SERIALIZED_VARIABLES.update({
        "external_flow_source_depth": float
    })

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.external_flow_source_depth = None

    def get_external_flow_source_depth(self):
        # If an external_flow_source_depth is not set, it is defaulted to the Riser bottom depth
        # This is only used for hydraulics, which will be implemented later.
        if self.external_flow_source_depth is None:
            self.external_flow_source_depth = self.bottom_depth

        return self.external_flow_source_depth



from typing import List, Union

from worker.app.modules import Module
from worker.app.modules.time_activity_module import TimeActivityModule
from worker.data.operations import gather_data_for_period
from worker.event import Event, EventType
from worker.event.scheduled import ScheduledEvent
from worker.event.stream import StreamEvent
from worker.framework import constants
from worker.mixins.logging import LoggingMixin
from worker.mixins.rollbar import RollbarMixin
from worker.state.mixins import RedisMixin
from worker.state.state import State
from worker import exceptions


class App(RedisMixin, LoggingMixin, RollbarMixin):
    """
    The apps is designed to receive the events of a single asset
    """
    app_state_fields = {
        'asset_id': int,
        'last_processed_timestamp': int
    }

    def __init__(self, *args, **kwargs):
        self.app_key = constants.get('global.app-key')
        self.app_name = constants.get('global.app-name')
        self.module_key = None

        self.event_type: EventType = None

        self.asset_id = None
        self.event = None  # event records

        super().__init__(*args, **kwargs)

    def load(self, event_type: EventType, event: Event):
        """
        :param event_type:
        :param event: a scheduler event or wits stream; belong to one asset
        :return:
        """
        self.event_type = event_type

        self.asset_id = event.asset_id

        self.state = self.load_state()

        max_lookback_seconds = self.get_max_lookback_seconds()
        event = self.load_event(event, max_lookback_seconds)

        valid_stream_collections = self.get_valid_stream_collections()
        self.event = self.filter_event_for_collections(self.event_type, valid_stream_collections, event)  # event records

        self.log_event(self.event, max_lookback_seconds)  # event records

    def log_event(self, event: List[dict], max_lookback_seconds: int):
        self.debug(self.asset_id, "WITS input to {0} -> {1}".format(self.app_name, event))

        if not event:
            return

        batch_size = len(event)
        start_time = event[0].get("timestamp")
        end_time = event[-1].get("timestamp")

        self.log(
            self.asset_id,
            text=f"Received {batch_size} elements from {start_time} to {end_time}. {max_lookback_seconds} seconds of initial data are lookback."
        )

    @staticmethod
    def get_valid_stream_collections():
        valid_collections = constants.get('global.valid-stream-collections', [])
        if not isinstance(valid_collections, (str, list)):
            raise TypeError("Incorrect type of valid-stream-collections in global constants")

        if isinstance(valid_collections, str):
            valid_collections = [valid_collections]
        return valid_collections

    def get_max_lookback_seconds(self):
        """
        For each module (mostly in time-base modules), the time of processing does not
        match the last time of the event so extra data is required to look back and get
        the data so the processing can start from where it left off.
        :return:
        """

        time_modules = [module for module in self.get_modules() if issubclass(module, TimeActivityModule)]
        maximum_lookback = 0
        for module in time_modules:
            module_lookback = constants.get('{}.{}.lookback-duration'.format(self.app_key, module.module_key), default=0)
            maximum_lookback = max(module_lookback, maximum_lookback)

        return maximum_lookback

    def load_event(self, event: Event, max_lookback_seconds: int) -> List[dict]:
        if self.event_type == EventType.SCHEDULER:
            return self.load_scheduler_event(self.asset_id, event, max_lookback_seconds)

        if self.event_type == EventType.WITS_STREAM:
            return self.load_wits_stream_event(self.asset_id, event, max_lookback_seconds)

        return None

    @staticmethod
    def filter_event_for_collections(event_type: EventType, valid_stream_collections: list, event: list) -> list:
        """
        This function filters the incoming event based on a list of valid collections

        :param event_type: Type of incoming event.
        :param valid_stream_collections: List of valid collections
        :param event: List of data records
        :return: List of records whose collection is one of the allowed event collections
        """

        # If event type is scheduler or event_collections is empty, return the entire event
        if event_type == EventType.SCHEDULER or not valid_stream_collections:
            return event

        # Filtering the event based on valid collections
        event = [record for record in event if record.get("collection") in valid_stream_collections]
        return event

    def filter_event_for_duplicates(self, event) -> list:
        last_processed_timestamp = self.state.get('last_processed_timestamp') or 0

        # If event is a single record and greater than the last processed timestamp, return the event
        if len(event) == 1 and event[0].get("timestamp") > last_processed_timestamp:
            return event

        # If length of unique timestamps is same as length of event and the first record timestamp is also greater than
        # last_processed_timestamp, then return the event
        unique_timestamps = set([record.get("timestamp") for record in event])
        if len(unique_timestamps) == len(event) and event[0].get("timestamp") > last_processed_timestamp:
            return event

        # Filtering the events for duplicates, once we identify that duplicates exist.
        filtered_events = []
        for each_record in event:
            if each_record.get("timestamp") > last_processed_timestamp:
                filtered_events.append(each_record)
                last_processed_timestamp = each_record.get("timestamp")

        return filtered_events

    def load_scheduler_event(self, asset_id: int, event: ScheduledEvent, max_lookback_seconds: int) -> List[dict]:
        """
        To load a scheduler event and get the wits stream data
        :param asset_id: The asset to load
        :param event: A cleaned event
        :param max_lookback_seconds: Maximum amount of time to look back prior to the scheduler event to cover gaps
        :return: list of WITS data between the last processed timestamp and the final event item timestamp
        """

        start_timestamp = self.state.get('last_processed_timestamp', event[0].start_time - 1)
        end_timestamp = event[-1].start_time

        # the event is converted from scheduler to wits stream
        scheduler_event = gather_data_for_period(
            asset_id=asset_id,
            start=start_timestamp - max_lookback_seconds,
            end=end_timestamp,
            limit=constants.get('global.query-limit'),
            fields=constants.get('global.wits_query_fields', None)
        )

        return scheduler_event

    def load_wits_stream_event(self, asset_id: int, event: StreamEvent, max_lookback_seconds: int) -> List[dict]:
        """
        To load a wits stream event and get more data if necessary
        :param asset_id: The asset to load
        :param event: A cleaned event
        :param max_lookback_seconds: Maximum amount of time to look back prior to WITS data to cover gaps
        :return: list of WITS data between the first event timestamp and the first timestamp
        """
        records = event.records

        # First filtering original event for duplicates
        records = self.filter_event_for_duplicates(records)

        # If all record timestamps are before the last processed timestamp,
        # the first timestamp will be the last processed timestamp from cache
        if not records:
            first_timestamp = self.state.get('last_processed_timestamp') or 0
        else:
            first_timestamp = records[0].get('timestamp')

        if not first_timestamp:
            return records

        if max_lookback_seconds:
            # Subtract one from the timestamp so that we don't reselect the final data item that was sent in the event
            end_timestamp = first_timestamp - 1

            previous_events = gather_data_for_period(
                asset_id=asset_id,
                start=first_timestamp - max_lookback_seconds,
                end=end_timestamp,
                limit=constants.get('global.query-limit'),
                fields=constants.get('global.wits_query_fields', None)
            )

            records = previous_events + records

        return records

    @staticmethod
    def determine_asset_id(event: list) -> int:
        try:
            return int(event[0]["asset_id"])
        except Exception:
            raise Exception('Event does not contain asset_id: {}'.format(event))

    def load_state(self, state_key: Union[str, None] = None, raise_warnings: bool = False) -> dict:
        previous_state = super().load_state(state_key=state_key, raise_warnings=raise_warnings)

        state = State(self.app_state_fields, previous_state)

        if not state.get('asset_id', None):
            state['asset_id'] = self.asset_id

        return state

    def get_modules(self) -> List[Module]:
        raise NotImplementedError("No modules found")

    def get_active_modules(self) -> List[Module]:
        return [module for module in self.get_modules() if module.enabled]

    def run_modules(self):
        if not self.event:  # event records
            return

        for module_type in self.get_active_modules():
            try:
                module = module_type(self.state, rollbar=self.rollbar)
            except Exception:
                raise exceptions.Misconfigured("Module {0} not able to initialize for asset_id {1}".format(module_type, self.asset_id))

            try:
                module.run(self.event)  # event records
            except Exception as ex:
                message = f"Error in module {module_type.module_key}"
                ex.args += (message,)  # Adding message to existing Exception
                raise

        last_processed_timestamp = self.event[-1].get('timestamp')  # event records
        self.state['last_processed_timestamp'] = last_processed_timestamp



from enum import Enum

from worker.data.serialization import serialization


class PipeMovementStatus(Enum):
    UNCLASSIFIED = -1
    IN_SLIPS = 0
    STATIC = 1
    ON_BOTTOM = 2
    PULL_OUT_OF_HOLE = 3
    RUN_IN_HOLE = 4
    DRILLING = 5

    def is_moving_out(self) -> bool:
        move_actions = [
            PipeMovementStatus.PULL_OUT_OF_HOLE,
        ]
        return self in move_actions

    def is_moving_in(self) -> bool:
        move_actions = [
            PipeMovementStatus.RUN_IN_HOLE,
            PipeMovementStatus.DRILLING
        ]
        return self in move_actions

    def is_moving(self) -> bool:
        return self.is_moving_in() or self.is_moving_out()


class PumpStatus(Enum):
    OFF = 0
    ON = 1


class RotationStatus(Enum):
    OFF = 0
    ON = 1


@serialization
class Activity(bytes, Enum):
    ROTARY_DRILLING = (
        "Rotary Drilling",
        PipeMovementStatus.DRILLING, PumpStatus.ON, RotationStatus.ON
    )
    SLIDE_DRILLING = (
        "Slide Drilling",
        PipeMovementStatus.DRILLING, PumpStatus.ON, RotationStatus.OFF
    )

    RUN_IN_HOLE = (
        "Run in Hole",
        PipeMovementStatus.RUN_IN_HOLE, PumpStatus.OFF, RotationStatus.OFF
    )
    WASHING_DOWN = (
        "Washing Down",
        PipeMovementStatus.RUN_IN_HOLE, PumpStatus.ON, RotationStatus.OFF
    )
    DRY_REAMING_DOWN = (
        "Dry Reaming Down",
        PipeMovementStatus.RUN_IN_HOLE, PumpStatus.OFF, RotationStatus.ON
    )
    REAMING_DOWN = (
        "Reaming Down",
        PipeMovementStatus.RUN_IN_HOLE, PumpStatus.ON, RotationStatus.ON
    )

    PULL_OUT_OF_HOLE = (
        "Pull out of Hole",
        PipeMovementStatus.PULL_OUT_OF_HOLE, PumpStatus.OFF, RotationStatus.OFF
    )
    WASHING_UP = (
        "Washing Up",
        PipeMovementStatus.PULL_OUT_OF_HOLE, PumpStatus.ON, RotationStatus.OFF
    )
    DRY_REAMING_UP = (
        "Dry Reaming Up",
        PipeMovementStatus.PULL_OUT_OF_HOLE, PumpStatus.OFF, RotationStatus.ON
    )
    REAMING_UP = (
        "Reaming Up",
        PipeMovementStatus.PULL_OUT_OF_HOLE, PumpStatus.ON, RotationStatus.ON
    )

    STATIC_OFF_BOTTOM = (
        "Static Off Bottom",
        PipeMovementStatus.STATIC, PumpStatus.OFF, RotationStatus.OFF
    )
    CIRCULATING = (
        "Circulating",
        PipeMovementStatus.STATIC, PumpStatus.ON, RotationStatus.OFF
    )
    DRY_ROTARY_OFF_BOTTOM = (
        "Dry Rotary Off Bottom",
        PipeMovementStatus.STATIC, PumpStatus.OFF, RotationStatus.ON
    )
    ROTARY_OFF_BOTTOM = (
        "Rotary Off Bottom",
        PipeMovementStatus.STATIC, PumpStatus.ON, RotationStatus.ON
    )

    STATIC_ON_BOTTOM = (
        "Static On Bottom",
        PipeMovementStatus.ON_BOTTOM, PumpStatus.OFF, RotationStatus.OFF
    )
    CIRCULATING_ON_BOTTOM = (
        "Circulating On Bottom",
        PipeMovementStatus.ON_BOTTOM, PumpStatus.ON, RotationStatus.OFF
    )
    DRY_ROTARY_ON_BOTTOM = (
        "Dry Rotary On Bottom",
        PipeMovementStatus.ON_BOTTOM, PumpStatus.OFF, RotationStatus.ON
    )
    CIRCULATING_AND_ROTARY_ON_BOTTOM = (
        "Circ. & Rot. On Bottom",
        PipeMovementStatus.ON_BOTTOM, PumpStatus.ON, RotationStatus.ON
    )

    IN_SLIPS = (
        "In Slips",
        PipeMovementStatus.IN_SLIPS, PumpStatus.OFF, RotationStatus.OFF
    )

    UNCLASSIFIED = (
        "Unclassified",
        PipeMovementStatus.DRILLING, PumpStatus.OFF, RotationStatus.OFF
    )

    def __new__(cls, activity_name: str, pipe_movement_status, pump_status, rotation_status):
        obj = bytes.__new__(cls)

        obj._value_ = activity_name  # using the activity name as the main value
        obj.pipe_movement_status = pipe_movement_status
        obj.pump_status = pump_status
        obj.rotation_status = rotation_status

        return obj

    def is_pipe_moving(self) -> bool:
        return self.pipe_movement_status.is_moving()

    def is_pipe_moving_out(self) -> bool:
        return self.pipe_movement_status.is_moving_out()

    def is_pipe_moving_in(self) -> bool:
        return self.pipe_movement_status.is_moving_in()

    def is_pumping(self) -> bool:
        return self.pump_status == PumpStatus.ON

    def is_rotating(self) -> bool:
        return self.rotation_status == RotationStatus.ON

    def is_drilling(self) -> bool:
        return self.pipe_movement_status == PipeMovementStatus.DRILLING

    def is_reaming(self) -> bool:
        """
        Pipe moves up and down with rotation
        :return:
        """
        movement_status = self.pipe_movement_status in [
            PipeMovementStatus.PULL_OUT_OF_HOLE,
            PipeMovementStatus.RUN_IN_HOLE
        ]
        rotation_status = self.rotation_status == RotationStatus.ON
        return movement_status and rotation_status

    def is_tripping(self) -> bool:
        """
        Pipe moves up and down
        :return:
        """
        if self == self.RUN_IN_HOLE or self == self.PULL_OUT_OF_HOLE:
            return True
        return False

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.value



from typing import Union, List

from worker.data.operations import get_data_by_path, get_config_by_id, compare_float

from worker import API
from worker.exceptions import MissingConfigError

from worker.wellbore.model.drillstring import Drillstring
from worker.wellbore.model.enums import HoleType
from worker.wellbore.model.hole import Hole
from worker.wellbore.model.hole_section import HoleSection
from worker.wellbore.model.riser import Riser
from worker.wellbore.wellbore import Wellbore

MIN_LENGTH = 1  # ft


def run_drillstring_and_create_wellbore(wits: dict, mud_flow_in: float = None) -> Union[Wellbore, None]:
    """
    Get the wits record and create a wellbore
    :param wits: wits records in json format
    :param mud_flow_in: used to determine if under-reamer needs to be activated for hole enlargement
    :return: a wellbore
    """
    if wits is None:
        return None

    api_worker = API()

    drillstring_id = get_data_by_path(wits, "metadata.drillstring", str, default=None)

    if not drillstring_id:
        return None

    drillstring = get_config_by_id(drillstring_id, collection="data.drillstring")

    if not drillstring:
        return None

    active_drillstring_number = drillstring.get("data", {}).get("id")

    # setting active drillstring
    drillstring = Drillstring(drillstring)

    bit = drillstring.get_bit()
    if not bit:
        raise MissingConfigError(f"Bit was not found in the drillstring with _id='{drillstring_id}'")

    asset_id = wits['asset_id']
    bit_size = bit.size

    bit_depth = get_data_by_path(wits, "data.bit_depth", float)
    hole_depth = get_data_by_path(wits, "data.hole_depth", float)

    # setting cased-hole sections
    query = "{data.inner_diameter#gte#%s}" % bit_size
    casings = api_worker.get(path="/v1/data/corva", collection="data.casing", asset_id=asset_id, query=query, limit=100).data
    hole = set_casings(casings)
    smallest_cased_hole_diameter = hole.get_min_inner_diameter() or 100

    # ====== setting open-hole sections
    query = "{data.id#lte#%s}" % active_drillstring_number
    sort = "{data.id:1}"
    drillstrings = api_worker.get(path="/v1/data/corva", collection="data.drillstring", asset_id=asset_id, query=query, sort=sort, limit=100).data
    hole = add_open_holes(hole, drillstrings, smallest_cased_hole_diameter, hole_depth)

    mud_flow_in = get_mud_flow_in(mud_flow_in, wits)

    return Wellbore(
        drillstring=drillstring,
        hole=hole,
        bit_depth=bit_depth,
        hole_depth=hole_depth,
        mud_flow_in=mud_flow_in,
    )


def run_casingstring_and_create_wellbore(wits: dict, mud_flow_in: float = None, is_design: bool = False) -> Union[Wellbore, None]:
    """
    Get the wits record and create a wellbore
    :param wits: wits records in json format
    :param mud_flow_in: used to determine if under-reamer needs to be activated for hole enlargement
    :param is_design: the mode of function
    :return: a wellbore
    """
    if wits is None:
        return None

    api_worker = API()

    casingstring_id = get_data_by_path(wits, "metadata.casing", str, default=None)

    if not casingstring_id:
        return None

    casing_data_source = "design.data.casing" if is_design else "data.casing"
    casingstring = get_config_by_id(casingstring_id, collection=casing_data_source)

    if not casingstring:
        return None

    casing_outer_diameter = casingstring.get("data", {}).get("outer_diameter")
    casing_inner_diameter = casingstring.get("data", {}).get("inner_diameter")
    casing_top_depth = casingstring.get("data", {}).get("top_depth")
    casing_bottom_depth = casingstring.get("data", {}).get("bottom_depth")
    casing_linear_weight = casingstring.get("data", {}).get("linear_weight")
    casing_components = casingstring.get("data", {}).get("components")

    # For best results in running a casing/liner in the well, the components should be presented.
    # If not we create an alternative.
    if not casing_components:
        # If it is a casing that reaches to the surface, we can use the provided information on the top to construct a casing string.
        casing_components = [
            {
                "family": "casing_joints",
                "linear_weight": casing_linear_weight,
                "length": casing_bottom_depth - casing_top_depth,
                "outer_diameter": casing_outer_diameter,
                "inner_diameter": casing_inner_diameter
            }
        ]

        # For liners, we use the top components for the casing joint part of the liner.
        # From the surface to the liner hanger, we create the same component with the family of DP (it is not the actual casing part and will be removed later).
        if casing_top_depth > Wellbore.DEPTH_THRESHOLD:
            add_component = {
                "family": "dp",
                "linear_weight": casing_linear_weight,
                "length": casing_top_depth,
                "outer_diameter": casing_outer_diameter,
                "inner_diameter": casing_inner_diameter
            }
            casing_components.insert(0, add_component)

        casingstring['data']['components'] = casing_components

    # setting active drillstring
    casingstring = Drillstring(casingstring)

    asset_id = wits['asset_id']

    bit_depth = get_data_by_path(wits, "data.bit_depth", float)
    hole_depth = get_data_by_path(wits, "data.hole_depth", float)

    query = "{data.inner_diameter#gt#%s}" % casing_inner_diameter
    sort = "{data.inner_diameter:-1}"
    previous_casings = api_worker.get(
        path="/v1/data/corva",
        collection=casing_data_source,
        asset_id=asset_id,
        query=query,
        sort=sort,
        limit=100
    ).data

    drilling_data_source = "data.well-sections" if is_design else "data.drillstring"
    sort = "{data.diameter:-1}" if is_design else "{timestamp: -1}"

    drill_sections = api_worker.get(
        path="/v1/data/corva",
        collection=drilling_data_source,
        asset_id=asset_id,
        sort=sort,
        limit=100
    ).data

    # use data.well-sections to create hole
    if is_design:
        hole = create_hole_using_well_sections_data(
            previous_casings,
            drill_sections,
            casing_bottom_depth
        )

    # use data.drillstring to create hole
    else:
        hole = create_hole_use_drillstring_data(
            previous_casings,
            drill_sections,
            casing_outer_diameter,
            hole_depth
        )

    mud_flow_in = get_mud_flow_in(mud_flow_in, wits)

    return Wellbore(
        drillstring=casingstring,
        hole=hole,
        bit_depth=bit_depth,
        hole_depth=hole_depth,
        mud_flow_in=mud_flow_in,
    )


def get_detailed_hole_from_elements(elements: List[dict],
                                    previous_element_top_depth: float,
                                    possible_top_depth: float,
                                    extend_casing_to_top: bool,
                                    is_riser: bool
                                    ) -> Hole:
    """
    This function converts elements of casing/riser components into a Hole.
    :param elements: list casing/riser components
    :param previous_element_top_depth: will be used as the bottom depth of current component
    :param possible_top_depth:  possible top depth to which a casing/riser can extend. In case of a casing, it will extend to the btm of riser (if available)
    :param extend_casing_to_top: if the casing/riser should extend to the top
    :param is_riser: If its a riser or a cased hole
    :return: Hole with all the components as HoleSections
    """
    casing_hole: Hole = Hole()

    elements = [element for index, element in enumerate(elements) if check_casing_element_validity(element, index, extend_casing_to_top)]

    for element in reversed(elements):
        element_id = element.get("inner_diameter", 0.0)
        element_length = element.get("length", 0.0)
        element_bottom = previous_element_top_depth
        element_top = element_bottom - element_length

        if element_top < possible_top_depth:
            create_section_and_append(casing_hole, element_id, possible_top_depth, element_bottom, HoleType.CASED_HOLE, is_riser)
            break

        create_section_and_append(casing_hole, element_id, element_top, element_bottom, HoleType.CASED_HOLE, is_riser)
        previous_element_top_depth = element_top

    casing_hole.sections.reverse()

    # checking and making sure the top element extends to the possible top depth.
    # This check is needed because we are ignoring components of dp family which maybe at the top of the liner.
    if compare_float(casing_hole.sections[0].top_depth, possible_top_depth, tolerance=0.01) != 0:
        casing_hole.sections[0].set_top_depth(possible_top_depth)
        casing_hole.sections[0].set_length()

    return casing_hole


def create_section_and_append(casing_hole, inner_diameter, top_depth, bottom_depth, hole_type, is_riser):
    sec = HoleSection(inner_diameter=inner_diameter, top_depth=top_depth, bottom_depth=bottom_depth, hole_type=hole_type)
    if is_riser:
        sec = Riser(inner_diameter=inner_diameter, top_depth=top_depth, bottom_depth=bottom_depth, hole_type=hole_type)

    casing_hole.sections.append(sec)


def check_casing_element_validity(casing_element: dict, idx: int, extend_casing_to_top: bool) -> bool:
    """
    Check if the component is a casing and of a significant length
    :param casing_element:
    :param idx:
    :param extend_casing_to_top:
    :return:
    """
    element_length: float = casing_element.get("length")
    element_family: str = casing_element.get("family")

    if not element_family or not isinstance(element_family, str):
        return False

    # Ignoring casing components with length less than 20 feet and only components that are not top most or those that don't extend to top depth.
    if element_length < 20.0 and (idx != 0 or not extend_casing_to_top):
        return False
    # Only accept the casing joints and riser components
    return True if element_family in ["casing_joints", "riser"] else False


def set_casings(casings: List[dict]) -> Hole:
    """
    Set the casing by providing the list of the jsons
    :param casings: list of casings from API in json dict format
    :return: None
    """
    if not casings:
        return Hole()

    possible_casing_top_depth: float = 0.0
    riser = next((csg for csg in casings if csg.get('data', {}).get('is_riser', False)), {})

    if riser:
        possible_casing_top_depth = riser.get("data").get("bottom_depth")
        casings.remove(riser)

    # sorting the casings in the ascending order (the smaller the inner_diameter the earlier it is added)
    casings = sorted(casings, key=lambda csg: csg.get('data', {}).get('inner_diameter'))

    # finding if a liner is overlapping outer liners
    # this is common in offshore drilling where they set the casings at the sea floor
    index = 1
    while index < len(casings):
        top_depth_inner = casings[index - 1].get('data', {}).get('top_depth', 1000)
        top_depth_outer = casings[index].get('data', {}).get('top_depth', 1000)
        if top_depth_outer + Wellbore.DEPTH_THRESHOLD > top_depth_inner:
            casings.pop(index)
        else:
            index += 1

    # reversing the order (the larger the inner_diameter the earlier it is)
    casings = reversed(casings)

    casing_holes = []
    for index, casing in enumerate(casings):
        casing_hole = Hole()
        casing_object: dict = casing.get("data", {})
        casing_bottom = casing_object.get("bottom_depth")
        casing_top = possible_casing_top_depth
        current_casing_top_depth = casing_object.get("top_depth")

        # The top depth of the first component should reach to surface regardless
        # Case 1 - Land (Normal)
        # Case 2 - Riser
        # Case 3 - Riserless
        if not (index == 0 and current_casing_top_depth < Wellbore.DEPTH_THRESHOLD):
            casing_top = current_casing_top_depth

        casing_elements: List[dict] = casing_object.get("components", [])
        if not casing_elements:
            hole_section = HoleSection(**casing_object, hole_type=HoleType.CASED_HOLE)
            casing_hole.sections.append(hole_section)
        else:
            previous_casing_element_top_depth = casing_bottom
            # Individual casing elements have to be looped in reverse order to correctly account for the topmost element going to surface is not liner
            extend_casing_to_top: bool = False
            if index == 0:
                extend_casing_to_top = True
            casing_hole = get_detailed_hole_from_elements(casing_elements, previous_casing_element_top_depth, casing_top, extend_casing_to_top, False)

        casing_holes.append(casing_hole)

    # Now adding the components of the riser to casingHoles
    if riser:
        riser_elements: List[dict] = riser.get("data", {}).get("components", [])
        previous_riser_element_top_depth = possible_casing_top_depth
        riser_hole: Hole = get_detailed_hole_from_elements(riser_elements, previous_riser_element_top_depth, 0.0, True, True)
        casing_holes.insert(0, riser_hole)

    return Hole.merge_holes(casing_holes)


def add_open_holes(hole: Hole, drillstrings: List[dict], smallest_cased_hole_diameter: float, hole_depth: float) -> Hole:
    for ds in drillstrings:
        bit = Drillstring(ds).get_bit()
        if not bit:
            drillstring_id = ds.get("_id")
            raise MissingConfigError(f"Bit was not found in the drillstring with _id='{drillstring_id}'")

        bit_size = bit.size
        # only keeping the DSs that ran below the smallest casing
        if bit_size > smallest_cased_hole_diameter:
            continue

        start_depth = ds.get("data", {}).get("start_depth")

        # if the open hole size increases then the prior open hole section IDs should be updated
        for section in reversed(hole):
            if section.hole_type == HoleType.OPEN_HOLE:
                if section.inner_diameter < bit_size:
                    section.inner_diameter = bit_size
            else:
                break

        previous_open_hole_size = hole.get_open_hole()[-1].inner_diameter if len(hole.get_open_hole()) > 0 else None

        # if the previous drillstring section inner diameter is equal to this bit size
        # there is no need to add a new drillstring section
        if previous_open_hole_size and previous_open_hole_size == bit_size:
            continue

        # if due to the side track the hole depth reduced, then it should update the setting depths in the open hole
        hole.sections = [
            sec for sec in hole.sections
            if sec.hole_type != HoleType.OPEN_HOLE or sec.top_depth <= hole_depth
        ]

        # if a new open-hole section is created in addition to the previous one,
        # the previous bottom depth may need to be modified.
        previous_hole_section = hole.get_last_hole_section()

        top_depth = hole.get_bottom_depth()
        bottom_depth = top_depth + 0.0001
        current_hole_section = HoleSection(
            top_depth=top_depth,
            bottom_depth=bottom_depth,
            inner_diameter=bit_size,
            hole_type=HoleType.OPEN_HOLE
        )

        if previous_hole_section:
            # The casing bottom depth can't be modified if the difference is more than 300 ft
            cased_hole_condition = previous_hole_section.hole_type == HoleType.CASED_HOLE and start_depth - previous_hole_section.bottom_depth < 300
            if cased_hole_condition or previous_hole_section.hole_type == HoleType.OPEN_HOLE:
                previous_hole_section.set_bottom_depth(start_depth)
                previous_hole_section.set_length()

            if current_hole_section.eq_without_length(previous_hole_section):
                continue

        current_hole_section.set_length()
        hole.add_section(current_hole_section)

    # In some cases the DS setting depth doesn't match the hole depth so the wits data overrides it;
    # this is only applied to the top depth of the last section and bottom depth of the previous section.
    # Note: only in cases that the wits hole_depth < setting depth
    # Condition: the number of open hole sections should be at least two needs to be satisfied first
    if len(hole) - len(hole.get_cased_hole()) > 1 and 1 < hole_depth < hole[-1].top_depth:
        hole[-2].set_bottom_depth(hole_depth)
        hole[-2].set_length()

        hole[-1].set_top_depth(hole_depth)
        hole[-1].set_length()

    return hole


def get_mud_flow_in(mud_flow_in: Union[float, None], wits: dict) -> Union[float, None]:
    """
    Get the mud flow rate from wits if `mud_flow_in` is None
    :param mud_flow_in:
    :param wits:
    :return:
    """
    if mud_flow_in is not None:
        return mud_flow_in

    if not wits:
        return None

    return get_data_by_path(wits, "data.mud_flow_in", default=None)


def create_hole_use_drillstring_data(
        previous_casings: List[dict],
        drillstrings: List[dict],
        casing_outer_diameter: float,
        hole_depth: float
) -> Union[Hole, None]:
    """
    Create a hole using data.drillstring collection
    :param previous_casings: previous casings data
    :param drillstrings: drillstring data
    :param casing_outer_diameter: outer diameter of casing in inch
    :param hole_depth: bottom depth of the hole in ft
    :return: a hole
    """
    drillstrings = [
        string for string in drillstrings
        if (string.get("data", {}).get("components", [{}])[-1].get("size") or -1) > casing_outer_diameter
    ]

    hole = set_casings(previous_casings)
    smallest_cased_hole_diameter = hole.get_min_inner_diameter() or 100

    return add_open_holes(hole, drillstrings, smallest_cased_hole_diameter, hole_depth)


def create_hole_using_well_sections_data(
        previous_casings: List[dict],
        drill_sections: List[dict],
        casing_bottom_depth: float
) -> Union[Hole, None]:
    """
    Create a hole using data.well-sections collection
    :param previous_casings: previous casings data
    :param drill_sections: well sections data
    :param casing_bottom_depth: bottom depth of casing in ft
    :return: a hole
    """
    # setting cased-hole sections
    hole = Hole()
    previous_casing_bottom_depth = 0
    if previous_casings:
        previous_casing_bottom_depth = previous_casings[-1].get("data", {}).get("bottom_depth")
        hole.set_casings(previous_casings)

    # setting open-hole sections
    for section in drill_sections:
        section_data = section.get("data", {})
        open_hole_top_depth = section_data.get("top_depth")
        open_hole_bottom_depth = section_data.get("bottom_depth")
        open_hole_diameter = section_data.get("diameter")

        if previous_casing_bottom_depth <= open_hole_top_depth <= open_hole_bottom_depth <= casing_bottom_depth + Wellbore.DEPTH_THRESHOLD:
            open_hole = HoleSection(
                inner_diameter=open_hole_diameter,
                top_depth=open_hole_top_depth,
                bottom_depth=open_hole_bottom_depth,
                hole_type=HoleType.OPEN_HOLE)
            hole.add_section(open_hole)

    return hole



import simplejson as json

from worker import API
from worker.data.json_encoder import JsonEncoder
from worker.mixins.logging import LoggingMixin
from worker.mixins.rollbar import RollbarMixin


class Alert(LoggingMixin, RollbarMixin):
    """
    Alert class can be used to trigger alerts from a data app.
    An event based alert has to be created on the Corva UI with a unique identifier.

    How to use:
    1. Create an instance of Alert with the same unique identifier and asset_id.
    2. Call trigger_alert and pass the required parameters to trigger the alert.
    3. To trigger an alert on Corva, the well must be active and visible to the user. (Current rules)
    """
    def __init__(self, asset_id: int, identifier: str, last_trigger_timestamp=None, *args, **kwargs):
        self.asset_id = asset_id
        self.identifier = identifier
        self.last_trigger_timestamp = last_trigger_timestamp

        super().__init__(*args, **kwargs)

    def trigger_alert(self, timestamp: int, timestamp_read: int, context: dict):
        """
        Function used to trigger the alert for the unique identifier.

        :param timestamp: timestamp at which the event for the alert is detected
        :param timestamp_read: timestamp at which the wits record was read by the source app
        :param context: A dict of data to be sent to the alerts engine
        :return:
        """

        context = context or {}
        context.update({
            "identifier": self.identifier,
            "asset_id": self.asset_id,
            "timestamp": timestamp,
            "timestamp_read": timestamp_read
        })

        api = API()
        try:
            trigger = api.post(
                "/v1/alerts/definitions/trigger/", data=json.dumps(context, cls=JsonEncoder, ignore_nan=True)
            ).data
            self.debug(self.asset_id, f"Triggered alert with context -> {context} and received response {trigger}")
            self.last_trigger_timestamp = timestamp

            return trigger
        except Exception as ex:
            message = f"Error while triggering alert for context {context} with exception {ex}"
            self.debug(self.asset_id, f"Failed to trigger alert with context -> {context}")
            self.track_error(message)


class Alerter:
    alerter = None
    asset_id = None

    @classmethod
    def set_asset_id(cls, asset_id):
        cls.asset_id = asset_id

    @classmethod
    def get_alerter(cls):
        if cls.alerter is None:
            cls.alerter = Alert(0, "")

        return cls.alerter

    @classmethod
    def trigger_alert(cls, identifier: str, message: str, timestamp: int = None, timestamp_read: int = None, context: dict = None):
        if not cls.asset_id:
            return

        context = context or {}
        context.update({
            "message": message
        })

        cls.get_alerter().asset_id = cls.asset_id
        cls.get_alerter().identifier = identifier
        cls.get_alerter().trigger_alert(timestamp, timestamp_read, context)



from typing import Union

import simplejson as json

from worker.data.api import API
from worker.data.json_encoder import JsonEncoder
from worker.framework import constants
from worker.mixins.logging import LoggingMixin
from worker.mixins.rollbar import RollbarMixin
from worker.state.mixins import RedisMixin
from worker.state.state import State


class Module(RedisMixin, LoggingMixin, RollbarMixin):
    """
    This is an abstract base module that needs to be extended by an actual module.
    """
    # module_key is used for redis access and state of this module
    module_key = "module"
    collection = "collection"
    module_state_fields = {
        'last_processed_timestamp': int
    }

    enabled = True

    def __init__(self, global_state, *args, **kwargs):
        self.app_key = constants.get('global.app-key')
        self.app_name = constants.get('global.app-name')
        self.asset_id = global_state.get("asset_id")

        self.global_state = global_state

        super().__init__(*args, **kwargs)

    def run(self, wits_stream: list):
        """
        :param wits_stream: a wits stream event
        :return:
        """
        # load the state
        state = self.get_state()
        self.state = self.process_module_state(state)

        # subclasses should implement their own run

    def should_run_processor(self, event):
        raise Exception("This method need to be implemented by subclasses!")

    def get_state(self, state_key: [str, None] = None, raise_warnings: bool = False) -> dict:
        current_state = super().load_state(state_key=state_key, raise_warnings=raise_warnings)
        return State(self.module_state_fields, current_state)

    @staticmethod
    def process_module_state(state):
        return state

    def set_state(self, key, value):
        self.state[key] = value

    def load_dataset(self, event):
        return event

    def run_module(self, dataset: Union[list, dict], beginning_results: list) -> list:
        raise Exception("Not implemented")

    def get_last_exported_timestamp_from_collection(self, asset_id, query=None, less_than=None):
        """
        Query the module collection for this asset_id + module, sorted by timestamp descending,
        limit 1, grab the last item's timestamp. Default to 0 if no records found.
        @asset_id:
        @less_than: the timestamp before which you want to get
        """
        if less_than:
            query = query or ""
            query += "AND{timestamp#lt#%s}" % less_than

        worker = API()
        last_document = worker.get(
            path="/v1/data/corva", query=query, collection=self.collection, asset_id=asset_id,
            sort="{timestamp: -1}", limit=1,
        ).data

        if not last_document:
            return 0

        last_document = last_document[0]
        last_processed_timestamp = last_document.get('timestamp', 0)

        return last_processed_timestamp

    @staticmethod
    def gather_first_wits_timestamp_since(asset_id: int, since: int, activity_fields=None, operator='eq') -> int:
        """
        Query the Wits collection for this asset_id where state in wits_states and timestamp >= since
        """

        query = '{timestamp#%s#%s}' % ('gt', since)

        operator = operator.lower()

        if activity_fields:
            if operator == "eq" and isinstance(activity_fields, list):
                operator = "in"

            if operator in ("in", "nin"):
                if not isinstance(activity_fields, list):
                    activity_fields = [activity_fields]

                # Put each state into a formatted string for querying
                activity_fields = ["'{0}'".format(state) for state in activity_fields]
                activity_fields = "[{0}]".format(",".join(activity_fields))
            else:
                activity_fields = "'{0}'".format(activity_fields)

            query += 'AND{data.state#%s#%s}' % (operator, activity_fields)

        worker = API()
        first_wits_since = worker.get(
            path="/v1/data/corva", collection='wits', asset_id=asset_id, sort="{timestamp: 1}", limit=1, query=query
        ).data

        if not first_wits_since:
            return 0

        first_wits_since = first_wits_since[0]
        first_wits_since_timestamp = first_wits_since.get('timestamp', 0)

        return first_wits_since_timestamp

    @staticmethod
    def gather_maximum_timestamp(event, start, activity_fields):
        """
        get the maximum time stamp of a stream of data
        :param event: a stream of data  that the majority is wits collection
        :param start:
        :param activity_fields:
        :return:
        """
        maximum_timestamp = start
        for data in event:
            if data.get("collection") == "wits" and data.get('data', {}).get('state', None) in activity_fields:
                maximum_timestamp = max(data.get("timestamp", 0), maximum_timestamp)

        return maximum_timestamp

    def gather_minimum_timestamp(self, asset_id: int, event: list):
        minimum = self.get_last_exported_timestamp_from_collection(asset_id)

        if not minimum:
            minimum = event[0]["timestamp"] - 1800

        return minimum

    def gather_collections_for_period(self, asset_id, start, end, query=None):
        limit = constants.get("global.query-limit")

        query = query or ""
        if query:
            query += "AND"

        query += "{timestamp#gte#%s}AND{timestamp#lte#%s}" % (start, end)

        worker = API()
        dataset = worker.get(
            path="/v1/data/corva", collection=self.collection, asset_id=asset_id, query=query,
            sort="{timestamp: 1}", limit=limit,
        ).data

        if not dataset:
            return []

        return dataset

    def store_output(self, asset_id, output):
        """
        to store/post results
        :param asset_id: asset id of the well
        :param output: an array of json objects to be posted
        :return: None
        """

        if not asset_id or not output or not self.collection:
            return

        output = self.format_output(output)

        self.debug(asset_id, "{0} output -> {1}".format(self.module_key, output))

        worker = API()
        worker.post(path="/v1/data/corva", data=output)

    def build_empty_output(self, wits: dict) -> dict:
        """
        Building an empty output result.
        :param wits: one wits record
        :return:
        """
        output = {
            'timestamp': int(wits.get('timestamp')),
            'company_id': int(wits.get('company_id')),
            'asset_id': int(wits.get('asset_id')),
            'provider': str(wits.get('provider', 'corva')),
            'version': 1,
            'collection': self.collection,
            'data': {}
        }

        enable_output_type_field = constants.get(
            "{}.{}.enable-output-type-field".format(self.app_key, self.module_key), False
        )

        if not enable_output_type_field:
            return output

        # If a specific module enforces type field in the output collection,
        # build_empty_output adds an additional type field to the output.
        # This can be enforced by setting "enable-output-type-field" to True in the module app constants.
        output.update({'type': self.module_key})

        return output

    @staticmethod
    def format_output(output):
        output = json.dumps(output, cls=JsonEncoder, ignore_nan=True)
        return output



import abc
import simplejson as json
from enum import Enum
from functools import cached_property

from worker.data.json_encoder import JsonEncoder
from worker.framework import constants


class EventType(Enum):
    SCHEDULER = "scheduler"
    WITS_STREAM = "wits_stream"


class Event(abc.ABC):
    """An event class that holds the events of a single asset_id."""
    def __init__(self, event_type: EventType):
        self.event_type: EventType = event_type

        self.asset_id: int = None
        self.records = []
        self.app_connection_id = None

    def add_records(self, new_records: list) -> None:
        self.records.extend(new_records)

    @abc.abstractmethod
    def complete_event(self, api) -> None:
        raise NotImplementedError

    def __len__(self):
        return len(self.records)

    def __getitem__(self, index: int):
        return self.records[index]

    @cached_property
    def is_posting_to_message_producer(self) -> bool:
        """
        Whether to post to message producer for the wits stream and scheduler events or not.
        If there is no lambda app following your app then this should be false,
        because this process is intracting with the API which slows the process down.
        """
        is_posting: bool = constants.get('global.post-to-message-producer', False)
        return is_posting

    @abc.abstractmethod
    def build_message_producer_payload(self) -> dict:
        raise NotImplementedError

    def post_to_message_producer(self, api):
        if not self.is_posting_to_message_producer or not self.app_connection_id:
            return

        payload = self.build_message_producer_payload()
        api.post(path='/v1/message_producer', data=json.dumps(payload, cls=JsonEncoder, ignore_nan=True))



import json
import os
import sys
from functools import cached_property
from typing import List, Union

from worker.mixins.logging import LoggingMixin
from worker.mixins.rollbar import RollbarMixin
from worker.state import RedisHandler


class RedisMixin(LoggingMixin, RollbarMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.state = {}

    @cached_property
    def _size_limits(self):
        state_storage_limit_fatal = os.getenv("STATE_STORAGE_CRITICAL_LIMIT", 1_000.0)  # 1 MB
        state_storage_limit_warning = os.getenv("STATE_STORAGE_WARNING_LIMIT", 100.0)  # 100 kB
        size_limits = {
            "fatal": float(state_storage_limit_fatal),
            "warning": float(state_storage_limit_warning)
        }
        return size_limits

    def load_state(self, state_key: Union[str, None] = None, raise_warnings: bool = False) -> dict:
        """
        Load the state from redis
        :param state_key:
        :param raise_warnings: Set to True if size limits warnings should be printed
        :return:
        """
        if not state_key:
            state_key = self.get_formatted_state_key(self.asset_id, self.app_key, self.module_key)

        state = RedisHandler.load_state(state_key)
        size_object = self.check_state_size(state, state_key, raise_warnings=raise_warnings)

        self.debug(self.asset_id, f"Retrieved state of size {size_object} kb")
        self.state = state
        return state

    def save_state(self, state_key: Union[str, None] = None, raise_warnings: bool = True, **kwargs) -> None:
        """
        Save the state to redis
        :param state_key:
        :param raise_warnings: Set to True if size limits warnings should be printed
        :return:
        """
        if not state_key:
            state_key = self.get_formatted_state_key(self.asset_id, self.app_key, self.module_key)

        size_object = self.check_state_size(self.state, state_key, raise_warnings=raise_warnings)
        RedisHandler.save_state(self.state, state_key, **kwargs)
        self.debug(self.asset_id, f"Saved state of size {size_object} kb")

    def delete_states(self, state_keys: Union[List[str], str]):
        """
        Delete state for current module
        :param state_keys:
        :return:
        """
        if not isinstance(state_keys, list):
            state_keys = [state_keys]

        RedisHandler.delete_states(state_keys)
        self.debug(None, "Deleted state from Redis")

    @staticmethod
    def get_formatted_state_key(asset_id: int, app_key: str, module_key: Union[str, None] = None) -> str:
        """
        Returns the state key in Corva naming format
        :param asset_id:
        :param module_key:
        :param app_key:
        :return:
        """
        state_key = "corva/{0}.{1}".format(asset_id, app_key)

        if module_key:
            return "{0}.{1}".format(state_key, module_key)

        return state_key

    def check_state_size(self, state, state_key, raise_warnings=True):
        """
        Check the size of the state dictionary and generate warnings if necessary
        :param state:
        :param state_key:
        :param raise_warnings:
        :return:
        """
        size_object = sys.getsizeof(json.dumps(state)) / 1024

        if not raise_warnings:
            return size_object

        size_limit = self._size_limits["fatal"]
        if size_object > size_limit:
            message = f"State_key {state_key} is of size {size_object} kb > {size_limit} kb."
            self.fatal(self.asset_id, message)
            self.track_message(message, level="critical")
            return size_object

        size_limit = self._size_limits["warning"]
        if size_object > size_limit:
            message = f"State_key {state_key} is of size {size_object} kb > {size_limit} kb."
            self.warn(self.asset_id, message)
            return size_object

        return size_object



from copy import deepcopy
from typing import Union, List

from worker.data.operations import get_data_by_path
from worker.data.serialization import serialization
from worker.wellbore.model.enums import HoleType
from worker.wellbore.model.hole_section import HoleSection
from worker.wellbore.sections_mixin import SectionsMixin


@serialization
class Hole(SectionsMixin):
    DEPTH_THRESHOLD = 50  # ft
    LENGTH_THRESHOLD = 1  # ft

    SERIALIZED_VARIABLES = {
        'sections': list
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sections = kwargs.get('sections') or []

    def add_section(self, section: HoleSection):
        """
        Add a new hole section to the end of the hole.
        :param section:
        :return:
        """
        if not self:
            top_depth = section.top_depth

            if top_depth < Hole.DEPTH_THRESHOLD:
                section.top_depth = 0

        if self:
            section.top_depth = self[-1].bottom_depth

        # bottom depth is the source of truth, so we are updating the length
        section.set_length()

        self.sections.append(section)

    def insert_section(self, section: HoleSection, index: int):
        """
        Insert a new hole section in the given index.
        :param section:
        :param index:
        :return:
        """
        if index is not None:
            if self:
                section.top_depth = self[index - 1].bottom_depth
            section.set_bottom_depth()

            self.sections.insert(index, section)
            self.update_depths(start_index=index + 1)

    def insert_section_after(self, section: HoleSection, section_before: HoleSection):
        index_section_before = self.sections.index(section_before)
        self.insert_section(section, index_section_before + 1)

    def get_bottom_depth(self) -> float:
        if not self:
            return 0
        return self[-1].bottom_depth

    def update_depths(self, start_index: int = 0):
        """
        To update the top and bottom depth of the components
        :param start_index:
        :return:
        """
        if start_index < 1:
            self[0].top_depth = 0
            self[0].set_bottom_depth()
            start_index = 1

        for i in range(start_index, len(self)):
            self[i].top_depth = self[i - 1].bottom_depth
            self[i].set_bottom_depth()

    def get_cased_hole(self):
        """
        Get the cased sections of the hole.
        :return:
        """
        cased_hole = Hole()

        for sec in self:
            if sec.hole_type == HoleType.CASED_HOLE:
                cased_hole.add_section(sec)

        return cased_hole

    def get_open_hole(self):
        """
        Get the open sections of the hole.
        :return:
        """
        open_hole = Hole()

        for sec in self:
            if sec.hole_type == HoleType.OPEN_HOLE:
                open_hole.add_section(sec)

        return open_hole

    def get_casing_bottom_depth(self):
        return self.get_cased_hole().get_bottom_depth()

    def is_casing_exist(self) -> bool:
        """
        Get the cased sections of the hole.
        :return:
        """
        return any(sec.hole_type == HoleType.CASED_HOLE for sec in self)

    def update(self, hole_depth: Union[dict, float]) -> bool:
        """
        Update the last open hole segment of the hole.
        :param hole_depth: a wits dict or actual hole depth
        :return: if any update occurred or not
        """

        # If there are no sections in the hole or the last hole is Open Hole, do not update hole and return
        if not self or self[-1].hole_type != HoleType.OPEN_HOLE:
            return False

        if isinstance(hole_depth, dict):
            hole_depth = get_data_by_path(hole_depth, "data.hole_depth", float)

        # if the hole depth < casing bottom depth, then ignore
        cased_hole_bottom_depth = self.get_cased_hole().get_bottom_depth()
        if hole_depth < cased_hole_bottom_depth:
            return False

        self[-1].set_bottom_depth(hole_depth)
        self[-1].set_length()

        return True

    def set_casings(self, casings: List[dict]):
        """
        Set the casing by providing the list of the jsons
        :param casings: list of casings from API in json dict format
        :return: None
        """
        # TODO we need to consider using 'data.components' if available to set the casings
        self.sections = []

        if not casings:
            return

        # sorting the casings in the ascending order (the smaller the inner_diameter the earlier it is added)
        casings = sorted(casings, key=lambda csg: csg.get('data', {}).get('inner_diameter'))

        # removing the outer casings at the surface
        surface_index = next(
            (
                idx for idx, csg in enumerate(casings)
                if csg.get('data', {}).get('top_depth', 1000) < Hole.DEPTH_THRESHOLD
            ), 0
        )
        casings = casings[0: surface_index + 1]

        # finding if a liner is overlapping outer liners
        # this is common in offshore drilling where they set the casings at the sea floor
        index = 1
        while index < len(casings):
            top_depth_inner = casings[index - 1].get('data', {}).get('top_depth', 1000)
            top_depth_outer = casings[index].get('data', {}).get('top_depth', 1000)
            if top_depth_outer + Hole.DEPTH_THRESHOLD > top_depth_inner:
                casings.pop(index)
            else:
                index += 1

        # reversing the order (the larger the inner_diameter the earlier it is)
        casings = reversed(casings)

        for csg in casings:
            hole_section = HoleSection(**csg.get('data', {}), hole_type=HoleType.CASED_HOLE)
            self.add_section(hole_section)

    def update_casings(self, casings: 'Hole'):
        """
        To remove the casings of the current hole and
        replace them with the given one
        :param casings:
        :return:
        """
        casings = casings.get_cased_hole()
        # TODO: check if deepcopy is required
        open_holes = deepcopy([sec for sec in self if sec.hole_type == HoleType.OPEN_HOLE])

        self.sections = []
        self.sections.extend(casings)
        self.sections.extend(open_holes)

        self.update_depths()

    def get_min_inner_diameter(self) -> Union[float, None]:
        if not self:
            return None
        return self[-1].inner_diameter

    def get_last_hole_section(self):
        if self:
            return self[-1]
        return None

    def compute_get_total_volume(self, top_depth: float = None, bottom_depth: float = None) -> float:
        """
        Compute the volume of the hole section
        :return: volume of the hole section in FOOT^3
        """
        if top_depth is None and bottom_depth is None:
            return sum(sec.compute_get_volume() for sec in self)

        top_depth = top_depth or 0
        bottom_depth = bottom_depth or self.get_bottom_depth()

        total_volume = 0

        for index, section in enumerate(self):
            if section.bottom_depth <= top_depth:
                continue
            if section.top_depth >= bottom_depth:
                break

            area = section.compute_get_area()  # in INCH^2
            length = min(bottom_depth, section.bottom_depth) - max(top_depth, section.top_depth)  # in FOOT
            volume = area / 144 * length
            total_volume += volume

        return total_volume

    def trim_to_bottom_depth(self, depth):
        if not self.sections:
            return

        self.sections = [section for section in self.sections if section.top_depth <= depth]
        if self.sections:
            self.sections[-1].set_bottom_depth(depth)
            self.sections[-1].set_length()

            if self.sections[-1].length < Hole.LENGTH_THRESHOLD:
                self.sections.pop(-1)

    def trim_to_top_depth(self, depth):
        if not self.sections:
            return

        self.sections = [section for section in self.sections if section.bottom_depth > depth]
        if self.sections:
            self.sections[0].set_top_depth(depth)
            self.sections[0].set_length()

            if self.sections[0].length < Hole.LENGTH_THRESHOLD:
                self.sections.pop(0)

    @staticmethod
    def merge_holes(casing_holes: List):
        """
        This method can be used to merge Holes. The overlapping sections are trimmed and updated.
        The mergedHole is the ultimate hole being generated starting from the bottom and keep adding
        elements (sections) to the top of it.

        :param casing_holes: casingHoles List of holes to be merged from surface to bottom;
        the wider casing is closer to surface (index 0), except for the riser that is always the first section.
        :return: a merged hole
        """
        if not casing_holes:
            return None

        merged_hole = deepcopy(casing_holes[-1])

        if len(casing_holes) == 1:
            return merged_hole

        for i in range(len(casing_holes) - 2, -1, -1):
            casing_hole = casing_holes[i]
            current_section = casing_hole.sections[-1]
            bottom_section = merged_hole.sections[0]

            # comparing the inner diameter of the current section vs the one below
            if current_section.inner_diameter > bottom_section.inner_diameter > 0:
                # normal situation where the casings are telescopic from surface
                casing_hole.trim_to_bottom_depth(bottom_section.top_depth)
            else:
                # in case there is a riser smaller than the casing below.
                merged_hole.trim_to_top_depth(current_section.bottom_depth)

            for j in range(len(casing_hole.sections) - 1, -1, -1):
                merged_hole.sections.insert(0, casing_hole.sections[j])

        return merged_hole



import json
from copy import deepcopy
from typing import Union

from worker import API
from worker.data.operations import get_data_by_path, nanround, get_config_by_id, is_stream_app

from worker.data.serialization import serialization
from worker.mixins.logging import Logger
from worker.wellbore.model.annulus import Annulus
from worker.wellbore.model.drillstring import Drillstring
from worker.wellbore.model.drillstring_components import UnderReamer
from worker.wellbore.model.enums import HoleType
from worker.wellbore.model.hole import Hole
from worker.wellbore.model.hole_section import HoleSection


@serialization
class Wellbore:
    DEPTH_THRESHOLD = 50  # ft

    SERIALIZED_VARIABLES = {
        'original_drillstring': Drillstring,
        'original_hole': Hole,
        'bit_depth': float,
        'hole_depth': float,
        'last_hole_depth': float,
        'last_non_decreasing_hole_depth': float,
        'under_reamer_alert_timestamp': float
    }

    def __init__(self, **kwargs):
        super().__init__()

        self.under_reamer: Union[UnderReamer, None] = None
        self.under_reamer_alert_timestamp: float = 0.0

        self.original_drillstring: Drillstring = kwargs.get('original_drillstring') or kwargs.get('drillstring')
        self.actual_drillstring: Drillstring = kwargs.get('actual_drillstring')

        # Explicitly checking if original_hole is not None, because bool of Hole with no sections evaluates to False
        self.original_hole: Hole = kwargs.get('original_hole') if kwargs.get('original_hole') is not None else kwargs.get('hole')
        self.actual_hole: Hole = kwargs.get('actual_hole')

        self.actual_annulus: Annulus = kwargs.get('actual_annulus')

        self.bit_depth: float = kwargs.get('bit_depth') or 0.0
        self.hole_depth: float = kwargs.get('hole_depth')

        self.last_hole_depth: float = kwargs.get('last_hole_depth') or self.hole_depth
        self.last_non_decreasing_hole_depth: float = kwargs.get('last_non_decreasing_hole_depth') or self.last_hole_depth

        mud_flow_in = kwargs.get("mud_flow_in") or 0.0

        self.update(mud_flow_in=mud_flow_in)

    @property
    def annulus(self):
        return self.actual_annulus

    def update(self, bit_depth: Union[float, dict] = None, hole_depth: float = None, mud_flow_in: float = None) -> None:
        """
        :param bit_depth: a wits dict (that will be used for both bit and hole depths) or actual bit depth
        :param hole_depth: actual hole depth value
        :param mud_flow_in: mud flow rate to check under-reamer activation condition
        :return:
        """
        if mud_flow_in is None:
            mud_flow_in = 0.0

        if self.original_drillstring is None:
            return

        if isinstance(bit_depth, dict):
            bit_depth = get_data_by_path(bit_depth, "data.bit_depth", float)
            hole_depth = get_data_by_path(bit_depth, "data.hole_depth", float)

        self.bit_depth = bit_depth or self.bit_depth or 0.0

        # a qc for the hole depth: if bit depth > hole depth override it with bit depth
        if hole_depth is None:
            hole_depth = self.hole_depth or self.original_hole.get_bottom_depth()

        hole_depth = max(hole_depth, self.bit_depth)
        self.hole_depth = hole_depth

        self.actual_drillstring = deepcopy(self.original_drillstring)
        self.actual_drillstring.update(self.bit_depth)

        self.last_non_decreasing_hole_depth = max(self.last_non_decreasing_hole_depth, self.hole_depth)

        self.actual_hole = deepcopy(self.original_hole)
        self.actual_hole.update(self.hole_depth)

        under_reamer: UnderReamer = self.actual_drillstring.get_under_reamer()
        if under_reamer:
            self.under_reamer = under_reamer
            self.under_reamer.alerted_at_timestamp = self.under_reamer_alert_timestamp
            self.add_under_reamer_sections(mud_flow_in)
            self.under_reamer_alert_timestamp = self.under_reamer.alerted_at_timestamp
            self.actual_hole.update(self.hole_depth)

        self.actual_annulus = Annulus(drillstring=self.actual_drillstring, hole=self.actual_hole)

        self.last_hole_depth = self.hole_depth

    def add_under_reamer_sections(self, mud_flow_rate):
        current_hole: HoleSection = self.actual_hole.find_section_at_measured_depth(self.under_reamer.top_depth + 0.0001)

        # Determine a valid start of reaming depth: either casing shoe or open hole top depth.
        if current_hole.hole_type == HoleType.CASED_HOLE:
            start_reaming_depth = current_hole.bottom_depth
        elif current_hole.inner_diameter < self.under_reamer.ur_opened_od:
            start_reaming_depth = current_hole.top_depth
        else:
            return

        activation_mode = self.under_reamer.activate_under_reamer(mud_flow_rate, start_reaming_depth, current_hole.hole_type)

        # If unactivated, do nothing
        if activation_mode == self.under_reamer.UnderReamerActivationType.UNACTIVATED:
            return

        # Activated for the first time
        if activation_mode == self.under_reamer.UnderReamerActivationType.MUD_FLOW_RATE or \
                activation_mode == self.under_reamer.UnderReamerActivationType.OPEN_HOLE:
            drillstring = self.get_and_update_drill_string()

            # Update the original drillstring to match the chopped drillstring under-reamer component
            self.original_drillstring = Drillstring(drillstring)

            # Update API only if running a stream app
            if is_stream_app():
                drillstring = self.update_drillstring_api(drillstring)
                # If api request failed do not enlarge
                if not drillstring:
                    return

                Logger.debug(f"Updated API with under-reamer activation at depth {self.under_reamer.ur_opened_depth}")
            else:
                Logger.info(f"Under-reamer was opened at {self.under_reamer.ur_opened_depth }. "
                            f"However, not updating the ur_opened_depth on drillstring on api because it was not triggered from Stream app")

        reamed_length = self.under_reamer.bottom_depth - self.under_reamer.ur_opened_depth
        # If under reamer is shallower than the opened depth do nothing
        if reamed_length <= 0:
            return

        # This property is used for enabling split flow
        self.under_reamer.set_opened(True)

        ur_opened_od = self.under_reamer.ur_opened_od
        current_hole = self.actual_hole.find_section_at_measured_depth(self.under_reamer.ur_opened_depth + 0.01)

        # ENLARGEMENT
        # Enlarge the hole from previous run
        if current_hole.inner_diameter == ur_opened_od and current_hole.top_depth == self.under_reamer.ur_opened_depth:
            current_hole.set_length(reamed_length)
            return

        # Create a new reamed hole
        bottom_hole = deepcopy(current_hole)
        current_hole.set_bottom_depth(self.under_reamer.ur_opened_depth)
        current_hole.set_length()
        bottom_hole.set_top_depth(self.under_reamer.ur_opened_depth)
        bottom_hole.set_length()

        under_reamer_hole = HoleSection(inner_diameter=ur_opened_od, hole_type=HoleType.OPEN_HOLE, top_depth=0.0, bottom_depth=0, length=0)
        under_reamer_hole.set_length(reamed_length)

        # If the current hole is zero no need to add a new enlarged section, enlarge the current hole
        if current_hole.length <= 0.0:
            current_hole.inner_diameter = ur_opened_od
            current_hole.bottom_depth = self.under_reamer.bottom_depth
            current_hole.set_length()
            self.actual_hole.insert_section_after(bottom_hole, current_hole)
            return

        # Add enlarged hole section between opened depth and current under-reamer depth
        self.actual_hole.insert_section_after(under_reamer_hole, current_hole)
        self.actual_hole.insert_section_after(bottom_hole, under_reamer_hole)
        self.actual_hole.update(self.actual_hole.get_bottom_depth())

    def update_drillstring_api(self, drillstring: dict) -> Union[dict, None]:
        try:
            path = "/v1/data/corva/data.drillstring"
            res = API().post(path=path, data=json.dumps(drillstring))
            return res.data
        except Exception as e:
            Logger.error(e)

        return

    def get_and_update_drill_string(self):
        mongo_id = self.original_drillstring.mongo_id
        drillstring = get_config_by_id(mongo_id, collection="data.drillstring")

        ur_opened_depth = self.under_reamer.ur_opened_depth

        components = drillstring.get("data", {}).get("components", [])
        # Modify stringJson and post to api
        under_reamer_component = next((component for component in components if component.get("family") == "ur"), {})
        if not under_reamer_component:
            return

        under_reamer_component["ur_opened_depth"] = ur_opened_depth

        return drillstring

    def update_casing(self, casings: Hole) -> None:
        """
        If there is any update in the casings
        :param casings:
        :return:
        """
        self.original_hole.update_casings(casings)
        self.actual_hole.update_casings(casings)
        self.update()

    def trim_after_hole_depth_reduction(self, hole_depth: float) -> None:
        self.last_non_decreasing_hole_depth = hole_depth
        self.original_hole = deepcopy(self.actual_hole)

        self.actual_hole.sections = [sec for sec in self.actual_hole.sections if sec.top_depth <= hole_depth]
        if self.actual_hole:
            self.actual_hole[-1].bottom_depth = hole_depth
            self.actual_hole[-1].set_length()

        self.update(hole_depth, hole_depth)

    def compute_get_drillstring_body_volume_change(self, from_bit_depth: float, to_bit_depth: float) -> float:
        """
        Compute the change in the drillstring body volume between two bit depths
        :param from_bit_depth: start bit depth
        :param to_bit_depth: end bit depth
        :return: volume in FOOT^3
        """
        ds_from = deepcopy(self.original_drillstring)
        ds_from.update(from_bit_depth)
        body_volume_from = ds_from.compute_get_body_volume()

        ds_to = deepcopy(self.original_drillstring)
        ds_to.update(to_bit_depth)
        body_volume_to = ds_to.compute_get_body_volume()

        return body_volume_to - body_volume_from

    def compute_get_drillstring_outside_volume_change(self, from_bit_depth: float, to_bit_depth: float) -> float:
        """
        Compute the change in the drillstring solid volume between two bit depths using OD
        :param from_bit_depth: start bit depth
        :param to_bit_depth: end bit depth
        :return: volume in FOOT^3
        """
        ds_from = deepcopy(self.original_drillstring)
        ds_from.update(from_bit_depth)
        outside_volume_from = ds_from.compute_get_outside_volume()

        ds_to = deepcopy(self.original_drillstring)
        ds_to.update(to_bit_depth)
        outside_volume_to = ds_to.compute_get_outside_volume()

        return outside_volume_to - outside_volume_from

    def __repr__(self):
        return f"Wellbore:\n" \
               f"===Bit Depth / Hole Depth = {nanround(self.bit_depth)} / {nanround(self.hole_depth)}\n" \
               f"===Drillstring:\n{self.actual_drillstring}\n" \
               f"===Hole:\n{self.actual_hole}\n" \
               f"===Annulus:\n{self.actual_annulus}\n\n" \
               f"===Given Drillstring:\n{self.actual_drillstring}\n"



import json
import os
from typing import List, Literal, Tuple, Union

import numpy as np

from worker.data import math
from worker.data.api import API
from worker.data.enums import Environment, EventTypes
from worker.data.wits import WITS
from worker.exceptions import NotFound


def gather_data_for_period(
        asset_id: int, start: int, end: int, limit: int = 1800,
        collection: str = "wits", fields: str = None
) -> list:
    """
    Get the wits data from API for an asset over an interval
    :param asset_id: asset id
    :param start: start timestamp
    :param end: end timestamp
    :param limit: count of the data
    :param collection: any collection
    :param fields: fields to be filtered
    :return: a list of wits data
    """
    if start >= end:
        return []

    query = '{timestamp#gte#%s}AND{timestamp#lte#%s}' % (start, end)
    worker = API()

    wits_dataset = worker.get(
        path="/v1/data/corva", collection=collection, asset_id=asset_id, sort="{timestamp: 1}", limit=limit,
        query=query, fields=fields).data

    if not wits_dataset:
        return []

    return wits_dataset


def get_one_data_record(asset_id: int, timestamp_sort: int = -1, collection: str = "wits") -> dict:
    """
    Get the first or last wits record of a given asset
    :param asset_id:
    :param timestamp_sort:
    :param collection:
    :return:
    """
    api_worker = API()
    data = api_worker.get(path='/v1/data/corva/', collection=collection, asset_id=asset_id,
                          sort="{timestamp:%s}" % timestamp_sort, limit=1).data
    if not data:
        return {}

    return data[0]


def delete_collection_data_of_asset_id(asset_id: int, collections: Union[str, list]):
    """
    Delete all the data of a collection for an asset id.
    :param asset_id:
    :param collections: a collection or a list of collections
    :return:
    """
    worker = API()

    if isinstance(collections, str):
        collections = [collections]

    for collection in collections:
        path = "/v1/data/corva/%s" % collection
        query = "{asset_id#eq#%s}" % asset_id

        # Looping a maximum of 2000 times. Tries to delete a maximum of 7.2 million records.
        for _ in range(2000):
            # Upon delete API returns the following response {"deleted_count": 10000}
            res = worker.delete(path=path, query=query, limit=3600, retry_count=1).data
            records_deleted = res.get("deleted_count") or 0

            # API currently deletes 10000 records at a time, if the deleted count is less than 3600, break loop
            if records_deleted < 3600:
                break


def point_main_envs(env: Literal['qa', 'staging', 'production', 'local']):
    """
    The purpose of this function is to point main environment variables to
    the provided environment. This method updates the following environment
    variables:
    - API_ROOT_URL
    - API_KEY
    - CACHE_URL

    :param env: the environment to point to
    :return:
    """
    if not env:
        return

    # validating the environment
    Environment(env)

    api_url = os.getenv(f"API_ROOT_URL_{env}") or os.getenv("API_ROOT_URL")
    api_key = os.getenv(f"API_KEY_{env}") or os.getenv("API_KEY")
    cache_url = os.getenv(f"CACHE_URL_{env}") or os.getenv("CACHE_URL")

    if not all([api_url, api_key, cache_url]):
        raise ValueError("Missing environment variables!")

    new_envs = {
        'API_ROOT_URL': api_url,
        'API_KEY': api_key,
        'CACHE_URL': cache_url
    }

    os.environ.update(new_envs)


def setup_api_worker(env: Literal['qa', 'staging', 'production', 'local'], app_name: str) -> API:
    """
    Set up the Corva API worker
    :param env: environment ['local', 'qa', 'staging', 'production']
    :param app_name:
    :return: api worker
    """
    # validating the environment
    Environment(env)

    api_url = os.getenv(f"API_ROOT_URL_{env}") or os.getenv("API_ROOT_URL")
    api_key = os.getenv(f"API_KEY_{env}") or os.getenv("API_KEY")

    options = {
        "api_url": api_url,
        "api_key": api_key,
        "app_name": app_name
    }
    api_worker = API(**options)

    return api_worker


def setup_redis_worker(env: Literal['qa', 'staging', 'production', 'local']):
    """
    Set up the Redis worker
    :param env: environment ['local', 'qa', 'staging', 'production']
    :return: redis worker
    """
    # validating the environment
    Environment(env)

    cache_url = os.getenv(f"CACHE_URL_{env}") or os.getenv("CACHE_URL")

    import redis
    redis_worker = redis.Redis.from_url(cache_url, decode_responses=True)

    return redis_worker


def get_config_by_id(string_id: str, collection: str) -> Union[dict, None]:
    """
    Get the drillstring or casingstring from API by providing mongodb _id
    :param string_id: mongodb _id of the drillstring
    :param collection:
    :return:
    """
    string = None
    try:
        string = API().get_by_id(
            path="/v1/data/corva/", collection=collection, id=string_id
        ).data
    except NotFound:
        pass

    return string


def is_number(data):
    """
    Check and return True if data is a number, else return False
    :param data: Input can be string, number or nan
    :return: True or False
    """
    try:
        data_cast = float(data)
        if data_cast >= 0 or data_cast <= 0:  # to make sure it is a valid number
            return True

        return False
    except ValueError:
        return False
    except TypeError:
        return False


def is_finite(data):
    """
    Check if the given data is a finite number
    Note that the string representation of a number is not finite
    :param data:
    :return: True or False
    """
    try:
        return is_number(data) and np.isfinite(data)
    except (TypeError, ValueError):
        return False


def is_int(s: str) -> bool:
    """
    To check if the given string is an integer or not
    :param s:
    :return:
    """
    try:
        int(s)
        return True
    except ValueError:
        return False


def is_null(value) -> bool:
    """
    Will return True if the value is null, None, any variant of -999.25

    :param value:
    :return:
    """
    value = to_number(value)
    if value is None:
        return True

    if value >= 0:
        return False

    value = str(value)
    for each in ["-", ".", "0"]:
        value = value.replace(each, "")

    if value == "99925":
        return True

    return False


def to_number(data):
    """
    Check and return if the data can be cast to a number, else return None
    :param data: Input can be string, number or nan
    :return: A numbers
    """
    if is_number(data):
        return float(data)

    return None


def none_to_nan(data):
    """
    If data is a list, return list with None replaced with nan.
    If data is None, return nan
    :param data:
    :return:
    """
    if isinstance(data, list):
        return [np.nan if e is None else e for e in data]

    if data is None:
        return np.nan

    return data


def get_data_by_path(data: dict, path: str, func=lambda x: x, **kwargs):
    """
    To find the path to a key in a nested dictionary.
    Note that none of the keys should end up in a list
    :param data:
    :param path: path to the final key; example of the paths are:
        'data.X.Y'
        'data.bit_depth'
    :param func: the type of the data (int, str, float, ...)
    :param kwargs: pass default value in case the path not found;
    note that None is an acceptable default
    :return:
    """
    has_default = 'default' in kwargs
    default = kwargs.pop('default', None)

    if not path:
        if has_default:
            return default

        raise KeyError("No key provided")

    keys = path.split('.')

    while keys:
        current_key = keys.pop(0)

        if current_key not in data:
            if has_default:
                return default

            raise KeyError("{0} not found in path".format(current_key))

        data = data.get(current_key)

    if data is None:
        return None

    return func(data)


def is_in_and_not_none(d: dict, key: str):
    """
    An structured way of getting data from a dict.
    :param d: the dictionary
    :param key:
    :return: True or False
    """
    if key in d.keys() and d[key] is not None:
        return True

    return False


def nanround(value, decimal_places=2):
    """
    Similar to python built-in round but considering None values as well
    :param value:
    :param decimal_places:
    :return:
    """
    if is_number(value):
        return round(value, decimal_places)

    return None


def merge_dicts(d1: dict, d2: dict) -> dict:
    """
    Merge two dictionaries
    Note: the 2nd item (d2) has a higher priority to write items with similar keys
    :param d1:
    :param d2:
    :return:
    """
    d = {**d1, **d2}
    return d


def equal(obj1: object, obj2: object, params: List[str]) -> bool:
    """
    To check if two objects are equal by comparing the given parameters.
    :param obj1:
    :param obj2:
    :param params:
    :return:
    """
    if type(obj1) is not type(obj2):
        return False

    return all(getattr(obj1, param) == getattr(obj2, param) for param in params)


def get_cleaned_event_and_type(event) -> Tuple[Union[list, dict], EventTypes]:
    """
    validate and flatten the events and organize the data into a desired format

    Task and generic events format is : dict => {}
    Scheduler events format is: list of list of dict => [[{}]]
    Kafka events format is: list of dict => [{}]
    The above formats can be used to determine the format

    :param event: a scheduler of kafka stream
    :return: event and event_type
    """

    if not event:
        raise ValueError("Empty events")

    if isinstance(event, (str, bytes, bytearray)):
        event = json.loads(event)

    if isinstance(event, dict):
        if "asset_id" in event:
            return event, EventTypes.GENERIC

        if "task_id" in event:
            return event, EventTypes.TASK

        raise TypeError("Missing task_id or asset_id keys in event")

    if not isinstance(event, list):
        raise TypeError("Event is not a list or a dict")

    first_event = event[0]
    if isinstance(first_event, list):
        if first_event[0] and "schedule_start" in first_event[0]:
            return event, EventTypes.SCHEDULER

        raise Exception("Missing scheduler_start key in scheduler event")

    elif isinstance(first_event, dict):
        # new kafka stream format: list of json objects, each with metadata and records
        # event = [{"metadata": { ... }, "records": [ ... ]}, {"metadata": { ... }, "records": [ ... ]}]
        return event, EventTypes.STREAM

    raise TypeError("Event is not either a scheduler or kafka consumer")


def compute_time_step(records: Union[List[WITS], List[dict], List[float]], percent=50) -> Union[int, None]:
    """
    Compute the time step of the wits records
    If the intention is to split the data based on their activities a higher percent
    is recommended (such as 99%). For other cases 50% might work.

    :param records: a list of wits records in WITS or dict format, or a list of timestamps
    :param percent: this is an important parameter to get the correct time step
    :return:
    """
    if len(records) <= 3:
        return None

    if not 0 <= percent <= 100:
        raise ValueError(f"percent ({percent}) is out of [0, 100] range.")

    if isinstance(records[0], WITS):
        timestamps = [rec.timestamp for rec in records]
    elif isinstance(records[0], dict):
        timestamps = [rec.get('timestamp') for rec in records]
    else:
        timestamps = records

    timestamps = [timestamp for timestamp in timestamps if is_finite(timestamp)]

    # a list of time steps for all the timestamps
    diffs = np.diff(timestamps)

    time_step = math.percentile(diffs, percent)
    if not time_step:
        return None

    return int(time_step)


def compare_float(value1, value2, tolerance):
    """
    Compare two values based on the given tolerance
    :param value1: first number
    :param value2: second number
    :param tolerance: tolerance of the comparison
    :return: +1 when value1 > value2 + tolerance, -1 when value2 > value1 + tolerance, 0 otherwise
    """
    if value1 > value2 + tolerance:
        return 1

    if value2 > value1 + tolerance:
        return -1

    return 0


def is_stream_app():
    lambda_name = os.getenv("AWS_LAMBDA_FUNCTION_NAME")
    if not lambda_name:
        return False

    return "task" not in lambda_name.lower()



from typing import Union

from worker.wellbore.model.ann import Ann
from worker.wellbore.model.drillstring_components import Pipe
from worker.wellbore.model.hole_section import HoleSection


class SectionsMixin(object):
    def __init__(self, *args, **kwargs):
        self.sections = []

    def __len__(self):
        return len(self.sections)

    def __getitem__(self, index):
        return self.sections[index]

    def __delitem__(self, index):
        del self.sections[index]

    def __bool__(self):
        return len(self) != 0

    def __repr__(self):
        return "\n".join(str(sec) for sec in self.sections)

    def compute_volume(
            self,
            top_depth: float,
            bottom_depth: float
    ) -> float:
        '''
        Compute the volume between two depths
        :param top_depth: in ft
        :param bottom_depth: in ft
        :return: volume in cuft
        '''
        volume = 0
        for element in self:
            if element.top_depth <= bottom_depth <= element.bottom_depth:
                volume += (bottom_depth - top_depth) * element.area / 144
                break

            elif element.top_depth <= top_depth <= element.bottom_depth:
                volume += (element.bottom_depth - top_depth) * element.area / 144
                top_depth = element.bottom_depth

        return volume

    def find_section_at_measured_depth(self, measured_depth: float,
                                       at_exact_measured_depth: bool = False) -> Union[HoleSection, Ann, Pipe, None]:
        """
        :param measured_depth:
        :param at_exact_measured_depth:
        :return:
        """
        if measured_depth > self[-1].bottom_depth:
            return None

        for sec in self:
            if sec.top_depth < measured_depth <= sec.bottom_depth:
                return sec

        if at_exact_measured_depth:
            return None

        return self[0]



from worker.app import TimeActivityModule


class DisableModule(TimeActivityModule):
    """
    The purpose of this module is to check the disable functionality
    by turing it off from the testing section.
    """
    module_key = "mse"
    collection = "another-module.collection"



from typing import Union, List

from worker.event import Event, EventType


class SingleScheduledEvent:
    def __init__(self, single_stream_event: dict):
        self.asset_id = single_stream_event.get('asset_id') or 0

        self.start_time = single_stream_event.get('schedule_start', 0) / 1000
        self.end_time = single_stream_event.get('schedule_end', 0) / 1000
        self.schedule_id = single_stream_event.get('schedule')
        self.app_connection_id = single_stream_event.get('app_connection')

    def complete_event(self, api) -> None:
        """Sets schedule as completed."""
        if not self.schedule_id:
            return

        api.post(path=f'/scheduler/{self.schedule_id}/completed')


class ScheduledEvent(Event):
    """A scheduled event class that holds the events of a single asset_id."""
    def __init__(self, new_records: Union[SingleScheduledEvent, List[SingleScheduledEvent]]):
        super().__init__(EventType.SCHEDULER)

        if new_records:
            self.add_records(new_records)

        if self.records:
            self.app_connection_id = self.records[-1].app_connection_id

    def add_records(self, new_records: Union[SingleScheduledEvent, List[SingleScheduledEvent]]):
        if isinstance(new_records, SingleScheduledEvent):
            new_records = [new_records]

        if not self.records:
            self.asset_id = new_records[0].asset_id

        super().add_records(new_records)

    def complete_event(self, api) -> None:
        """
        Two steps can happen for scheduler events.
        1. Sets schedule as completed; this happens for all the events.
        2. If another lambda function is following your lambda then it should send a message to message producer as well.

        Args:
            api (API):
        """
        # Step 1: mark all the events as completed.
        for each in self.records:
            each.complete_event(api)

        # Step 2: in case another lambda function following your lambda
        self.post_to_message_producer(api)

    def build_message_producer_payload(self) -> dict:
        data = [
            {
                'timestamp': self.records[-1].start_time
            }
        ]

        return {
            'app_connection_id': self.app_connection_id,
            'asset_id': self.asset_id,
            'data': data
        }



from typing import List, Union

from worker.data.operations import get_data_by_path
from worker.event import Event, EventType
from worker.framework import constants


class StreamEvent(Event):
    """A stream event class that holds the events of a single asset_id."""
    def __init__(self, event: dict, is_posting_to_message_producer: bool = False):
        super().__init__(EventType.WITS_STREAM)

        self.metadata = event.get('metadata') or {}
        self.records = event.get('records') or []

        self.asset_id = 0
        if self.records:
            self.asset_id = self.records[0].get('asset_id')

        app_key = constants.get('global.stream_app_key')
        self.app_connection_id = get_data_by_path(
            self.metadata,
            f'apps.{app_key}.app_connection_id',
            func=int,
            default=None
        )

    @classmethod
    def merge(cls, events: List['StreamEvent']) -> Union['StreamEvent', None]:
        """Merge stream events of the same asset id into one"""
        if not events:
            return None

        merged_event = events[0]
        for each in events[1:]:
            merged_event.add(each)

        return merged_event

    def add(self, other: 'StreamEvent'):
        if self.asset_id != other.asset_id:
            raise Exception(f"Events of different assets can not be merged; {self.asset_id} and {other.asset_id}!")

        self.add_records(other.records)

    def build_message_producer_payload(self) -> dict:
        return {
            'app_connection_id': self.app_connection_id,
            'asset_id': self.asset_id,
            'data': self.records
        }

    def complete_event(self, api) -> None:
        self.post_to_message_producer(api)



import os
import reprlib


class RollbarMixin(object):
    def __init__(self, *args, **kwargs):
        self.rollbar = kwargs.pop('rollbar', None)
        super().__init__(*args, **kwargs)

    def is_rollbar(self) -> bool:
        """
        To check if rollbar is available or not
        :return: if rollbar is available
        """
        return self.rollbar and self.rollbar.SETTINGS.get('enabled')

    def track_message(self, message: str, level: str):
        """
        To send a message to rollbar
        :param message:
        :param level: any of the following levels:
        ['critical', 'error', 'warning', 'info', 'debug', 'ignored']
        :return:
        """
        # Levels:
        if not self.is_rollbar():
            print(f"{level} - {message}")
            return

        level = level.lower()

        self.rollbar.report_message(message, level)

    def track_error(self, message: str = None):
        if not self.is_rollbar():
            raise

        self.rollbar.report_exc_info(extra_data=message, level='error')


def payload_handler(payload: dict, **kw) -> dict:  # kw is currently unused
    """
    This is a rollbar payload handler which will be called on every payload being sent to rollbar
    This handler has to be added to rollbar instance after init

    Rollbar already applies a shortening transform to the payload using reprlib and custom sizes defined in rollbar init
    But it somehow seems to miss nested lists which is a usual format in our apps using corva-worker-python
    This payload handler applies the same reprlib to the lists missed by rollbar's shortener

    Use like -> rollbar.events.add_payload_handler(payload_handler)
    To pause trimming set REDUCED_ROLLBAR_PAYLOAD to False or false in env variables

    :param payload: Payload automatically captured based on exc, args and locals
    :param kw:
    :return: Trimmed payload
    """
    reduced_rollbar_payload_flag = os.getenv("REDUCED_ROLLBAR_PAYLOAD", "true").lower()

    # If REDUCED_ROLLBAR_PAYLOAD is set to False or false in env variables, trimming will be paused
    if reduced_rollbar_payload_flag == "false":
        return payload

    try:
        payload_frames = payload["data"]["body"]["trace"].pop("frames", [])
    except KeyError:
        return payload

    # All the locals are nested inside frames.
    # Smartly payload reduce size
    for frame in payload_frames:
        local_vars = frame.get("locals", {})

        # Instantiating repr from reprlib. Using this to trim nested datasets
        _repr = reprlib.Repr()

        # Setting custom max sizes for dict and list
        setattr(_repr, "maxdict", 10)
        setattr(_repr, "maxlist", 2)

        for key, value in local_vars.items():
            # Applying trim to only dicts and lists.
            # If there are other nested data types, they will be trimmed based on default repr limits
            if isinstance(value, (dict, list)):
                local_vars[key] = _repr.repr(value)

    payload["data"]["body"]["trace"]["frames"] = payload_frames
    return payload



import time

from worker import Module


class TestRedis(Module):
    module_key = "redis-tester"
    collection = "None"

    module_state_fields = {
        'message': str,
    }

    # override
    def run_module(self, event, beginning_results):
        start_timer = time.time()

        self.reset_states()
        self.test_save_load()
        self.test_delete_states()
        self.state = {"message": f"This state is auto saved for {self.module_key}"}
        print(f"Redis tests completed in {int(time.time() - start_timer)} seconds.\n")

    def run(self, wits_stream: list):
        self.run_module(wits_stream, [])

    def reset_states(self):
        # Initialize and reset states
        print("Initialize test")
        asset_id = self.global_state["asset_id"]
        state_key = self.get_formatted_state_key(asset_id, self.app_key, self.module_key)
        self.delete_states(state_key)
        self.get_state()

    def test_save_load(self):
        # Save/load tests
        print("Save/load test")
        self.state = {"message": "This state is and example state"}
        self.save_state()
        state = self.get_state()
        assert self.state == state
        print("Save/load test completed")

    def test_delete_states(self):
        # Deletion Tests
        asset_id = self.global_state["asset_id"]

        modules = []
        for i in range(5):
            _module_key = f"delete.test{i}"
            module = TestRedis(self.global_state)
            module.module_key = _module_key
            state_key = module.get_formatted_state_key(asset_id, module.app_key, _module_key)
            state = {"message": f"Redis state with {state_key=}"}
            print(f"Set the {state=}")
            module.state = state
            module.save_state()
            modules.append(module)
            time.sleep(0.1)  # back-off to prevent blocking by redis

        # Delete Multiple
        state_keys = [
            self.get_formatted_state_key(asset_id, module.app_key, module.module_key)
            for module in modules
        ]

        multiple_state_keys = state_keys[0:4]
        print(f"Deleting following {multiple_state_keys=}")
        self.delete_states(multiple_state_keys)
        for i in range(4):
            _module_key = f"delete.test{i}"
            current_state = modules[i].get_state()
            assert current_state == {"message": None}

        # Delete Single
        single_state_key = state_keys[4]
        print(f"Deleting following single {single_state_key=}")
        self.delete_states(single_state_key)
        current_state = modules[4].get_state()
        assert current_state == {"message": None}



import itertools
import json
from typing import Dict, Union
from worker.data.api import API

from worker import App
from worker.event import Event, EventType
from worker.event.scheduled import SingleScheduledEvent, ScheduledEvent
from worker.event.stream import StreamEvent
from worker.exceptions import EventFormatError
from worker.framework import constants


api = API()


class EventHandler:
    def __init__(self, app: App):
        self.app: App = app
        self.event_type: EventType = None
        self.event_by_asset_id: Dict[int, Event] = None

    def process(self, event: Union[str, dict]):
        """
        The whole process that is performed on an event including the
        loading, handling state, running, and completing the event.

        Args:
            event (Union[str, dict]): lambda handler event
        """
        self._load(event)
        self._run()

    def _load(self, event: Union[str, dict]):
        """
        Cleaning the event and group events based on their asset ids

        Args:
            event (Union[str, dict]): lambda handler event
        """
        self.event_type = self.get_event_type()
        event = self.clean_event(event)
        self.event_by_asset_id = self.format_event(self.event_type, event)

    def _run(self):
        """ Full run of the events based on their asset id """
        for asset_id, event in self.event_by_asset_id.items():
            self.app.load(self.event_type, event)
            self.app.run_modules()
            self.app.save_state()
            event.complete_event(api)

    @staticmethod
    def get_event_type() -> EventType:
        event_type: str = constants.get('global.event-type', "")
        return EventType(event_type)

    @staticmethod
    def clean_event(event):
        if not event:
            raise Exception("Empty event")

        if isinstance(event, (str, bytes, bytearray)):
            event = json.loads(event)

        return event

    @staticmethod
    def format_event(event_type: EventType, event: list) -> dict:
        """
        validate the wits_stream event, flatten and organize the data into a desired format
        :param event_type: type of event
        :param event: the wits or scheduler json event
        :return: a dict of records that are grouped by the asset_ids
        """
        if event_type == EventType.SCHEDULER:
            if not isinstance(event[0], list):
                raise EventFormatError("Invalid event!")

            # Scheduler events type is 'list of lists'; flattening into a single list
            events = [SingleScheduledEvent(item) for sublist in event for item in sublist]
            merging_function = ScheduledEvent

        else:  # 'wits_stream'
            events = [StreamEvent(each) for each in event]
            merging_function = StreamEvent.merge

        # sorting is required otherwise we only capture the last group of each asset_id
        events.sort(key=lambda single_event: single_event.asset_id)
        groups = itertools.groupby(events, key=lambda single_event: single_event.asset_id)

        events_by_asset_id = {
            group: merging_function(list(dataset))
            for group, dataset in groups
        }

        return events_by_asset_id



from worker.data.api import API
from worker.framework import constants
from worker.app import App
from worker.app.modules import Module

__all__ = [
    'API',
    'constants',
    'App',
    'Module'
]



import json
import os
from typing import List, Union

import redis

from worker.mixins.logging import Logger


class RedisHandler:
    redis_ttl = os.getenv("REDIS_TTL", 90 * 24 * 3600)  # 90 days
    _redis_connection = None

    @classmethod
    def _get_redis_connection(cls) -> redis.client.Redis:
        """
        Setup redis and get connection if one does not exist already
        :return:
        """
        # Use previous connection if cached
        if cls._redis_connection:
            Logger.debug("Using existing connection")
            return cls._redis_connection

        # If connection does not exist, create a new one
        cls._redis_connection = cls._new_redis_connection()

        return cls._redis_connection

    @classmethod
    def _new_redis_connection(cls) -> redis.client.Redis:
        """
        Creates a new Redis connection
        :return:
        """
        cache_url = os.getenv("CACHE_URL", None)

        if not cache_url:
            raise Exception("redis key (CACHE_URL) not found in Environment Variables.")

        _redis_connection = redis.Redis.from_url(cache_url)

        if not _redis_connection:
            raise Exception(f"Could not connect to Redis with URL: {cache_url}")

        client_id = _redis_connection.client_id()
        Logger.log(f"Created a new connection with {client_id=}")
        return _redis_connection

    @classmethod
    def setup_existing_connection(cls, existing_connection: Union[redis.client.Redis, None]) -> redis.client.Redis:
        """
        Set the internal connection variable to existing connection provided in this function
        The returned output should be saved to a global variable if it is intended to cache the connection across lambda invokes
        :param existing_connection:
        :return:
        """
        cls._redis_connection = existing_connection
        return cls._get_redis_connection()

    @classmethod
    def load_state(cls, state_key: str) -> dict:
        """
        Load state from redis
        :return:
        """
        state = cls._get_redis_connection().get(state_key)
        if state:
            return json.loads(state)

        return {}

    @classmethod
    def save_state(cls, state: dict, state_key: str, **kwargs):
        """
        Save the state to redis. Uses redis_ttl from kwargs if passed, fallbacks to env var then default 90 days
        :param state:
        :param state_key:
        :return:
        """
        # Using redis ttl from kwargs if passed in.
        redis_ttl = kwargs.pop("redis_ttl", cls.redis_ttl)
        cls._get_redis_connection().set(state_key, value=json.dumps(state), ex=redis_ttl)

    @classmethod
    def delete_states(cls, state_keys: List[str]):
        """
        Delete the states corresponding to given keys
        :param state_keys: a list of redis keys
        :return:
        """
        cls._get_redis_connection().delete(*state_keys)



import numpy as np
from typing import Union, Dict

from worker.data.activity import Activity
from worker.data.enums import DataStatus, ChannelStatus
from worker.data import operations


class Channel(float):
    def __new__(cls, value=None, status=None):
        # If only a valid value is passed, data status is automatically set to VALID
        # If None is passed as value, then data status is automatically set to MISSING
        # If both value and data status are passed, then it uses the passed data status
        if value is None:
            return None

        if not isinstance(value, float):
            try:
                value = float(value)
            except ValueError or TypeError:
                return None

        if value is not None and not status:
            status = DataStatus.VALID

        ch = float.__new__(cls, value)
        ch._status = status
        return ch

    @property
    def status(self) -> DataStatus:
        """
        - Status can be valid of missing
        - Data status is only available for non None wits attributes
        - Missing represents a value that has been forward filled or interpolated

        :return: DataStatus
        """
        return self._status


class WITS:
    """
    - Create a wits object by WITS.set_wits(dataset: Union[list, dict])
    - Access any wits data attribute by wits.hole_depth
    - Access any data status of a wits data attribute by wits.hole_depth.status
    - Data status is only available for non None wits attributes
    """
    data = {
        "entry_at": int,
        "hole_depth": Channel,
        "bit_depth": Channel,
        "block_height": Channel,
        "hook_load": Channel,
        "weight_on_bit": Channel,
        "rop": Channel,
        "rotary_rpm": Channel,
        "rotary_torque": Channel,
        "mud_flow_in": Channel,
        "mud_flow_out_percent": Channel,
        "standpipe_pressure": Channel,
        "diff_press": Channel,
        "pump_spm_1": Channel,
        "pump_spm_2": Channel,
        "pump_spm_3": Channel,
        "pump_spm_4": Channel,
        "pump_spm_total": Channel,
        "trip_tank_volume_1": Channel,
        "trip_tank_volume_2": Channel,
        "active_pit_volume": Channel,
        "lag_depth": Channel,
        "state": Activity,
        "boost_pump_flow_in": Channel,
        "annular_back_pressure": Channel
    }
    metadata = {
        "drillstring": str
    }
    spm_channels = ["pump_spm_1", "pump_spm_2", "pump_spm_3", "pump_spm_4"]
    status = {}

    def __init__(self, wits_json: dict = None, **kwargs):
        if not wits_json:
            wits_json = {}

        # Serialize functionality in Worker unpacks all the attributes. If that's how the class is called, repack.
        if kwargs:
            wits_json.update(dict(kwargs))
            wits_json.pop("_wits_", None)

        self.wits_json = wits_json

        data = wits_json.get("data", {})
        data_raw = wits_json.get("data_raw", {})

        metadata = wits_json.get("metadata", {})
        status = wits_json.get("status", {})

        for key in self.data.keys():
            # If type is not Channel or if the value is None, just setattr
            if self.data.get(key) != Channel or data.get(key, None) is None:
                setattr(self, key, data.get(key, None))
                continue

            key_data_status = DataStatus.VALID.value
            # If key is in data_raw, there are 2 possibilities - Missing (Value is None or -999.25) or Overridden
            if key in data_raw:
                key_data_status = DataStatus.OVERRIDDEN.value
                # If key in data_raw and value is null (None or -999.25), then we just setattr None
                if operations.is_null(data_raw.get(key, None)):
                    setattr(self, key, data.get(key, None))
                    continue

            # If type is Channel and (key not in data_raw or (key in data_raw and value is not null))
            setattr(self, key, Channel(data.get(key), DataStatus(data.get(f"{key}_data_status", key_data_status))))

        for key in self.metadata.keys():
            setattr(self, key, metadata.get(key, None))
        for key in self.status.keys():
            setattr(self, key, status.get(key, None))

        setattr(self, "timestamp", wits_json.get("data").get("entry_at"))
        setattr(self, "timestamp_read", wits_json.get("timestamp_read"))

        if getattr(self, "pump_spm_total") is None:
            self.set_pump_spm_total()

        setattr(self, "total_trip_tank_volume", self.get_total_trip_tank_volume())

    def __getattr__(self, item):
        return None

    def __setattr__(self, key, value):
        """
        When trying to set an attribute it automatically tries to typecast the value.
        :param key: key
        :param value: value
        :return:
        """
        if key not in {**self.data, **self.status}.keys():
            self.__dict__[key] = value
            return

        data_type = {**self.data, **self.status}.get(key)
        try:
            if not isinstance(value, Channel):
                value = data_type(value)
            self.__dict__[key] = value
        except TypeError:
            self.__dict__[key] = None
        except ValueError:
            self.__dict__[key] = None

    def __iter__(self):
        yield self

    @classmethod
    def set_wits(cls, dataset: Union[list, dict]):
        """
        Class method to create WITS objects using a dict or a list of dicts

        :param dataset: wits record or a list of wits records
        :return: WITS object or a list of WITS objects
        """
        if isinstance(dataset, dict):
            return cls(dataset)

        if isinstance(dataset, list):
            return [cls(record) for record in dataset]

    def get_as_dict(self) -> [None, dict]:
        """
        Get the wits record as a dictionary

        :return: wits record as a dict
        """
        if not self.wits_json:
            return None

        self.wits_json.pop("data_raw", None)
        self.wits_json["data"] = {key: getattr(self, key) for key in self.data}
        self.wits_json["data"].update({
            f"{key}_data_status": getattr(self, key).status.value
            for key in self.data
            if self.data.get(key) == Channel and getattr(self, key) is not None
        })
        # convert state from Activity type to str
        self.wits_json["data"]["state"] = self.wits_json["data"]["state"].value
        self.wits_json["status"] = {key: getattr(self, key) for key in self.status}

        return self.wits_json

    def get_wits_data(self, columns=None) -> dict:
        """
        Get the data part of the wits record as a dictionary

        :param columns: Specify if you only want a select list of columns
        :return: data part of wits record as a dict
        """
        if not columns:
            data = {key: getattr(self, key) for key in self.data}
            return data

        data = {key: getattr(self, key) for key in columns}
        return data

    def spm_status(self, low_spm_threshold: float) -> [ChannelStatus, None]:
        """
        Get spm status. Compares total of all SPM channels against SPM threshold.

        :param low_spm_threshold: threshold
        :return: None if channels are unavailable, OFF if less than threshold and ON if greater
        """

        # If all SPM channels are empty, return None
        if not any(operations.is_number(getattr(self, spm_channel)) for spm_channel in self.spm_channels):
            return ChannelStatus.MISSING

        if getattr(self, "pump_spm_total") > low_spm_threshold:
            return ChannelStatus.ON

        return ChannelStatus.OFF

    def mud_flow_in_status(self, low_mud_flow_in_threshold: float) -> [ChannelStatus, None]:
        """
        Get flow rate status. Compares mud flow in channel against flow rate threshold.

        :param low_mud_flow_in_threshold: threshold
        :return: None if channel is unavailable, OFF if less than threshold and ON if greater
        """
        if not operations.is_number(self.mud_flow_in):
            return ChannelStatus.MISSING

        if self.mud_flow_in > low_mud_flow_in_threshold:
            return ChannelStatus.ON

        return ChannelStatus.OFF

    def check_channel_availability(self, required_wits_channels: list) -> bool:
        """
        This function checks if all the required channels are available in the wits record

        :param required_wits_channels: List of required channels
        :return: Boolean value. True if all the required channels are available, else False
        """
        return all(operations.is_in_and_not_none(self.get_wits_data(required_wits_channels), channel)
                   for channel in required_wits_channels)

    def set_pump_spm_total(self):
        """This function sets the pump_spm_total attribute value based on individual spm values"""

        pump_spm_total = sum(getattr(self, spm_channel)
                             for spm_channel in self.spm_channels
                             if (getattr(self, spm_channel) and getattr(self, spm_channel) >= 0))

        setattr(self, "pump_spm_total", pump_spm_total)

    def get_total_trip_tank_volume(self) -> float:
        """
        This function returns the total trip tank volume if more than one trip tanks are available

        :return: Total trip tank volume
        """
        ttk1 = self.trip_tank_volume_1 or 0
        ttk2 = self.trip_tank_volume_2 or 0

        return ttk1 + ttk2

    def validate_range(self, channel_range_map: Dict[str, Dict]):
        """
        This method can be used to validate the range of the given channels and apply the proper operation on the data
        :param channel_range_map: a dictionary of channel to operation and range ([min, max]) mapping
        example:
            {
                "mud_flow_out_percent": {"range": [0, 100], "operation": "clip"}
            }
            options for 'operation': ['clip', None]
        :return: None,
        """
        for channel, properties in channel_range_map.items():
            orig_value = getattr(self, channel, None)
            if orig_value is None:
                continue

            valid_range = properties.get('range')
            operation = properties.get('operation', None)

            new_value = None

            if operation == 'clip':
                new_value = np.clip(orig_value, *valid_range)
            elif operation is None:
                if valid_range[0] <= orig_value <= valid_range[-1]:
                    new_value = orig_value
            else:
                raise ValueError("Invalid operation provided!")

            setattr(self, channel, new_value)


def serialize(obj: Union[list, WITS]) -> Union[list, dict, None]:
    """
    This function is used to serialize a WITS object or a list of WITS objects.
    Use WITS.set_wits(dataset: Union[list, dict]) to deserialize.

    :param obj: WITS object(s) to serialize.
    :return: return a serialized object
    """
    if isinstance(obj, list):
        s_obj = [each.get_as_dict() for each in obj]
        return s_obj

    if isinstance(obj, WITS):
        s_obj = obj.get_as_dict()
        return s_obj

    return None



from worker.event.event_handler import EventHandler
from testcase.app.drilling_efficiency_app import DrillingEfficiency
from worker.state import RedisHandler

redis_connection = None


def lambda_handler(event, context):
    """
    This function is the main entry point of the AWS Lambda function
    :param event: a scheduler or kafka event
    :param context: AWS Context
    :return:
    """

    global redis_connection
    redis_connection = RedisHandler.setup_existing_connection(redis_connection)

    app = DrillingEfficiency()
    event_handler = EventHandler(app)
    event_handler.process(event)



from testcase.state_testing_app.state_modules import TestRedis
from worker import App


class TestingApp(App):
    # override
    def get_modules(self):
        return [
            TestRedis,  # Redis testing module
        ]

    def load(self, event_type, event):
        # Do not want to run the super().load method
        return



from collections import OrderedDict
import itertools
from typing import List, Union

from worker.data import operations
from worker.data.activity import Activity
from worker.data.math import split_zip_edges
from worker.data.serialization import serialization
from worker.data.wits import WITS


@serialization
class ActivityGroup:
    SERIALIZED_VARIABLES = {
        'activity': Activity,
        'start': int,
        'end': int
    }

    def __init__(self, activity: Activity, start: int, end: int, **kwargs):
        self.activity = activity
        if end >= start:
            self.start = start
            self.end = end

    @property
    def duration(self) -> int:
        """
        :return: duration in seconds
        """
        return self.end - self.start + 1

    @classmethod
    def merge(cls, grp1: 'ActivityGroup', grp2: 'ActivityGroup',
              max_duration_gap: int = None, resultant_activity: Activity = None) -> Union['ActivityGroup', None]:
        """
        Merging two groups of activities into one.
        :param grp1:
        :param grp2:
        :param max_duration_gap: allowed gap between groups
        :param resultant_activity: if provided it will be used as the output activity, otherwise the activity
        of both groups are checked and if they are only the same it will continue the merging
        :return:
        """
        min_start = min(grp1.start, grp2.start)
        max_end = max(grp1.end, grp2.end)

        if max_duration_gap is not None:
            max_start = max(grp1.start, grp2.start)
            min_end = min(grp1.end, grp2.end)

            # if the gap between two activities are greater than the threshold, return None
            if max_start - min_end > max_duration_gap:
                return None

        if not resultant_activity:
            if grp1.activity != grp2.activity:
                return None
            resultant_activity = grp1.activity

        return ActivityGroup(resultant_activity, min_start, max_end)

    @staticmethod
    def has_overlap(group1: 'ActivityGroup', group2: 'ActivityGroup') -> bool:
        """
        Checking if two activity groups has overlap regardless of their precedence
        :param group1:
        :param group2:
        :return:
        """
        if not all([group1, group2]):
            return False

        max_start = max(group1.start, group2.start)
        min_end = min(group1.end, group2.end)

        return max_start <= min_end

    def __eq__(self, other):
        if not isinstance(other, ActivityGroup):
            return False

        return operations.equal(self, other, list(self.SERIALIZED_VARIABLES.keys()))

    def __repr__(self):
        return f"{self.activity.value:<25}: {self.start}-{self.end} -> duration={self.duration:>4}"


"""
In 1-second data frequency sometimes some activities will be added that can be
part of the bounding activities. For instance, you might be Reaming Down and
for a second the bit depth is the same as the previous timestamp and the activity
is Rotary Off Bottom. This will then be grouped into Reaming Down group.
"""
NEUTRAL_ACTIVITIES = OrderedDict([
    (Activity.RUN_IN_HOLE, [Activity.STATIC_OFF_BOTTOM]),
    (Activity.PULL_OUT_OF_HOLE, [Activity.STATIC_OFF_BOTTOM]),
    (Activity.WASHING_DOWN, [Activity.CIRCULATING]),
    (Activity.WASHING_UP, [Activity.CIRCULATING]),
    (Activity.DRY_REAMING_DOWN, [Activity.DRY_ROTARY_OFF_BOTTOM]),
    (Activity.DRY_REAMING_UP, [Activity.DRY_ROTARY_OFF_BOTTOM]),
    (Activity.REAMING_DOWN, [Activity.ROTARY_OFF_BOTTOM]),
    (Activity.REAMING_UP, [Activity.ROTARY_OFF_BOTTOM]),

    (Activity.STATIC_OFF_BOTTOM, [Activity.RUN_IN_HOLE, Activity.PULL_OUT_OF_HOLE]),
    (Activity.CIRCULATING, [Activity.WASHING_DOWN, Activity.WASHING_UP]),
    (Activity.DRY_ROTARY_OFF_BOTTOM, [Activity.DRY_REAMING_DOWN, Activity.DRY_REAMING_UP]),
    (Activity.ROTARY_OFF_BOTTOM, [Activity.REAMING_DOWN, Activity.REAMING_UP]),
])


class ActivityGrouping:
    def __init__(
        self, min_group_duration: int = 3, apply_neutral_grouping: bool = True, apply_gap_filling: bool = True,
        merging_dict: dict = None, time_step=None
    ):
        """
        :param min_group_duration: minimum duration of each group
        :param apply_neutral_grouping:
        :param apply_gap_filling:
        :param merging_dict: A dictionary representing the activities to be merged. For instance in the following
        example the Pull Out of Hole, Washing Up and Reaming Up are merged into one group and the activity of the
        whole group is PULL_OUT_OF_HOLE.
        {
            Activity.PULL_OUT_OF_HOLE : [Activity.PULL_OUT_OF_HOLE, Activity.WASHING_UP, Activity.REAMING_UP],
            ...
        }
        """
        self.min_group_duration = min_group_duration
        self.apply_neutral_grouping = apply_neutral_grouping
        self.apply_gap_filling = apply_gap_filling
        self.merging_dict = merging_dict
        self.time_step = time_step

    def group(self, wits_records: List[WITS]) -> List[ActivityGroup]:
        """
        to categorize a stream of wits data into groups of similar activities
        :param wits_records: wits records
        :return:
        """
        timestamps = [wits.timestamp for wits in wits_records]
        activities = [wits.state for wits in wits_records]

        if self.apply_neutral_grouping:
            activities = perform_neutral_grouping(activities)

        if not self.time_step:
            self.time_step = operations.compute_time_step(timestamps, percent=95)

        unique_activities = unique_everseen(activities)
        groups = []
        for activity in unique_activities:
            # Finding timestamps of matching activities
            valid_timestamps = [timestamps[i] for i in range(len(timestamps)) if activities[i] == activity]
            # Grouping and finding the edges of the activities
            edges = split_zip_edges(
                valid_timestamps,
                separation_length=self.time_step,
                min_segment_length=self.min_group_duration
            )
            this_groups = [ActivityGroup(activity=activity, start=start, end=end) for (start, end) in edges]
            groups.extend(this_groups)

        groups.sort(key=lambda x: x.start)

        if self.merging_dict:
            # 1. mapping the activities to their merging activities
            for grp in groups:
                grp.activity = self.merge_activity_map(grp.activity)

            # 2. merging adjacent activities with similar merging activities
            index = 1
            while index < len(groups):
                grp1, grp2 = groups[index - 1: index + 1]
                merged_group = ActivityGroup.merge(grp1, grp2, max_duration_gap=max(self.time_step, 2))
                if not merged_group:
                    index += 1
                    continue
                # update one and remove the next
                groups[index - 1] = merged_group
                groups.pop(index)

        if self.apply_gap_filling:
            groups = perform_gap_filling(groups)

        return groups

    def merge_activity_map(self, activity):
        """
        Map the constituent activity to the group activity
        :param activity:
        :return:
        """
        if not self.merging_dict:
            return activity

        for merging_group, constituent_activities in self.merging_dict.items():
            if activity in constituent_activities:
                return merging_group

        return activity


def perform_neutral_grouping(activities: List[Activity]):
    for bound_activity, neutral_activities in NEUTRAL_ACTIVITIES.items():
        for i in range(1, len(activities) - 1):
            if bound_activity == activities[i - 1] == activities[i + 1] and activities[i] in neutral_activities:
                activities[i] = bound_activity
    return activities


def perform_gap_filling(groups: List[ActivityGroup]):
    # TODO something to consider
    return groups


def unique_everseen(iterable, key=None):
    """
    from: https://docs.python.org/3.3/library/itertools.html
    List unique elements, preserving order. Remember all elements ever seen.
    :param iterable:
    :param key:
    :return:
    Example:
    unique_everseen('AAAABBBCCDAABBB') --> A B C D
    unique_everseen('ABBCcAD', str.lower) --> A B C D
    """
    seen = set()
    if key is None:
        for element in itertools.filterfalse(seen.__contains__, iterable):
            seen.add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen.add(k)
                yield element
